#include "fff_array.h"

#include <stdlib.h>
#include <errno.h>


fff_array* fff_array_from_data_new( fff_datatype datatype, void* buf, 
				    size_t dimX, size_t dimY, size_t dimZ, size_t dimT,
				    size_t offX, size_t offY, size_t offZ, size_t offT )
{
  fff_array* thisone; 
  fff_array_ndims ndims = FFF_ARRAY_4D; 

  /* Instantiate the structure member */ 
  thisone = (fff_array*)calloc( 1, sizeof(fff_array) ); 
  if ( thisone==NULL )
    fff_error("fff_array_from_data_new", "Out of memory", ENOMEM); 
      
  /* Decrease the number of dimensions if applicable */ 
  if ( dimT == 1 ) {
    ndims = FFF_ARRAY_3D; 
    if ( dimZ == 1 ) {
      ndims = FFF_ARRAY_2D;
      if ( dimY == 1 ) 
	ndims = FFF_ARRAY_1D;
    }
  }
  thisone->ndims = ndims; 

  /* Set dimensions / offsets / voxel size */ 
  thisone->dimX = dimX; 
  thisone->dimY = dimY; 
  thisone->dimZ = dimZ; 
  thisone->dimT = dimT; 
  thisone->offsetX = offX; 
  thisone->offsetY = offY; 
  thisone->offsetZ = offZ; 
  thisone->offsetT = offT; 

  /* Set data type and point towards buffer */ 
  thisone->datatype = datatype; 
  thisone->data = buf; 
  thisone->owner = 0; 
  
  return thisone; 
}



/*

Creates a C-contiguous array. 

*/ 

fff_array* fff_array_new( fff_datatype datatype,
			  size_t dimX, size_t dimY, size_t dimZ, size_t dimT )
{
  fff_array* thisone; 
  size_t nvoxels = dimX*dimY*dimZ*dimT; 
  size_t aux, offX, offY, offZ, offT;
  
  /* Offset computation */ 
  offT = 1; 
  aux = dimT; 
  offZ = aux; 
  aux *= dimZ; 
  offY = aux; 
  aux *= dimY; 
  offX = aux; 

  /* Create a new image */ 
  thisone =  fff_array_from_data_new( datatype, NULL, 
				      dimX, dimY, dimZ, dimT,
				      offX, offY, offZ, offT ); 

  /* Gives ownership */ 
  thisone->owner = 1; 

  /* Allocate the image buffer */ 
  switch( datatype ) {

  case FFF_UCHAR:
    {
      unsigned char* buf = (unsigned char*)calloc( nvoxels, sizeof(unsigned char) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case FFF_SCHAR:
    {
      signed char* buf = (signed char*)calloc( nvoxels, sizeof(signed char) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case FFF_USHORT:
    {
      unsigned short* buf = (unsigned short*)calloc( nvoxels, sizeof(unsigned short) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case FFF_SSHORT:
    {
      signed short* buf = (signed short*)calloc( nvoxels, sizeof(signed short) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case FFF_UINT:
    {
      unsigned int* buf = (unsigned int*)calloc( nvoxels, sizeof(unsigned int) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case FFF_INT:
    {
      int* buf = (int*)calloc( nvoxels, sizeof(int) ); 
      thisone->data = (void*)buf;  
    }
    break;
  case FFF_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)calloc( nvoxels, sizeof(unsigned long int) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case FFF_LONG:
    {
      long int* buf = (long int*)calloc( nvoxels, sizeof(long int) ); 
      thisone->data = (void*)buf;  
    }
    break;
  case FFF_FLOAT:
    {
      float* buf = (float*)calloc( nvoxels, sizeof(float) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case FFF_DOUBLE:
    {
      double* buf = (double*)calloc( nvoxels, sizeof(double) ); 
      thisone->data = (void*)buf;  
    }
    break;
  default: 
    fff_error("fff_array_new", "Unrecognized image type", EINVAL );
    break; 

  }
  
  /* Report error if array has not been allocated */ 
  if ( thisone->data==NULL )
    fff_error("fff_array_new", "Out of memory", ENOMEM); 

  return thisone; 
}


void fff_array_delete( fff_array* thisone )
{
  if ( (thisone->owner) && (thisone->data != NULL) ) 
    free(thisone->data); 
  free(thisone); 
  return; 
}



#define IMAGE_SET(val)						\
  i = 0;							\
  while ( iter.uncomplete ) {					\
    buf[i] = val;						\
    i += fff_array_iterator_update(&iter); }								

void fff_array_set_all( fff_array* thisone, double val )
{
  char* proc = "fff_array_set_all"; 
  fff_array_iterator iter = fff_array_iterator_init( thisone ); 
  size_t i;
  
  switch( thisone->datatype ) {

  case FFF_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case FFF_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case FFF_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      IMAGE_SET(val);
    }
    break;
  case FFF_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case FFF_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case FFF_INT:
    {
      int* buf = (int*)thisone->data;
      IMAGE_SET(val);  
    }
    break;
  case FFF_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case FFF_LONG:
    {
      long int* buf = (long int*)thisone->data;
      IMAGE_SET(val);  
    }
    break;
  case FFF_FLOAT:
    {
      float* buf = (float*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case FFF_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  default: 
    fff_warning( proc, "Unrecognized image type.\n" );
    break; 
  }

  return; 
}


/* Return FFF_NAN if position is out of bounds */ 
double fff_array_get( const fff_array* thisone, 
		      size_t x, 
		      size_t y, 
		      size_t z, 
		      size_t t )
{
  double value = 0.0; 
  char* proc = "fff_array_get"; 
  size_t offX = thisone->offsetX; 
  size_t offY = thisone->offsetY; 
  size_t offZ = thisone->offsetZ; 
  size_t offT = thisone->offsetT; 

  /* Check coordinate range */ 
  if ( (x >= thisone->dimX) ||
       (y >= thisone->dimY) ||
       (z >= thisone->dimZ) ||
       (t >= thisone->dimT) ) 
	return FFF_NAN; 


  switch(thisone->datatype) {

  case FFF_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case FFF_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case FFF_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf);
    }
    break;
  case FFF_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case FFF_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case FFF_INT:
    {
      int* buf = (int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case FFF_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case FFF_LONG:
    {
      long int* buf = (long int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case FFF_FLOAT:
    {
      float* buf = (float*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case FFF_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  default: 
    fff_warning( proc, "Unrecognized image type.\n" );
    break; 
  }
  
  return value; 
} 


void fff_array_set( fff_array* thisone, 
		    size_t x, 
		    size_t y, 
		    size_t z, 
		    size_t t, 
		    double value )
{
  char* proc = "fff_array_set"; 
  size_t offX = thisone->offsetX;
  size_t offY = thisone->offsetY; 
  size_t offZ = thisone->offsetZ; 
  size_t offT = thisone->offsetT; 

  /* Warning: coordinate range is not checked! */ 

  switch(thisone->datatype) {

  case FFF_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (unsigned char)(FFF_ROUND(value));
    }
    break;
  case FFF_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (signed char)(FFF_ROUND(value));
    }
    break;
  case FFF_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (unsigned short)(FFF_ROUND(value));
    }
    break;
  case FFF_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (signed short)(FFF_ROUND(value)); 
    }
    break;
  case FFF_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (unsigned int)(FFF_ROUND(value));
    }
    break;
  case FFF_INT:
    {
      int* buf = (int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (int)(FFF_ROUND(value)); 
    }
    break;
  case FFF_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (unsigned long int)(FFF_ROUND(value));
    }
    break;
  case FFF_LONG:
    {
      long int* buf = (long int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (long int)(FFF_ROUND(value)); 
    }
    break;
  case FFF_FLOAT:
    {
      float* buf = (float*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (float)value;
    }
    break;
  case FFF_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (double)value;
    }
    break;
  default: 
    fff_warning( proc, "Unrecognized image type.\n" );
    break; 
  }
  
  return; 
} 


double fff_array_get_from_index( const fff_array* thisone, 
				 size_t pos )
{
  double value = 0.0; 
  char* proc = "fff_array_get_from_index"; 

  /* Warning: coordinate range is not checked! */ 

  switch(thisone->datatype) {

  case FFF_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      buf += pos;
      value = (double)(*buf); 
    }
    break;
  case FFF_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      buf += pos;
      value = (double)(*buf); 
    }
    break;
  case FFF_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      buf += pos;
      value = (double)(*buf);
    }
    break;
  case FFF_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      buf += pos;
      value = (double)(*buf); 
    }
    break;
  case FFF_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      buf += pos;
      value = (double)(*buf);
    }
    break;
  case FFF_INT:
    {
      int* buf = (int*)thisone->data;
      buf += pos;
      value = (double)(*buf); 
    }
    break;
  case FFF_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      buf += pos;
      value = (double)(*buf);
    }
    break;
  case FFF_LONG:
    {
      long int* buf = (long int*)thisone->data;
      buf += pos;
      value = (double)(*buf); 
    }
    break;
  case FFF_FLOAT:
    {
      float* buf = (float*)thisone->data;
      buf += pos; 
      value = (double)(*buf); 
    }
    break;
  case FFF_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      buf += pos; 
      value = (double)(*buf); 
    }
    break;
  default: 
    fff_warning( proc, "Unrecognized image type.\n" );
    break; 
  }

  return value; 
}


void fff_array_set_from_index( fff_array* thisone, 
			       size_t pos, 
			       double value )
{
  char* proc = "fff_array_set_from_index"; 

  /* Warning: coordinate range is not checked! */ 

  switch(thisone->datatype) {

  case FFF_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      buf += pos;
      *buf = (unsigned char)(FFF_ROUND(value)); 
    }
    break;
  case FFF_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      buf += pos;
      *buf = (signed char)(FFF_ROUND(value)); 
    }
    break;
  case FFF_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      buf += pos;
      *buf = (unsigned short)(FFF_ROUND(value));
    }
    break;
  case FFF_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      buf += pos;
      *buf = (signed short)(FFF_ROUND(value));
    }
    break;
  case FFF_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      buf += pos;
      *buf = (unsigned int)(FFF_ROUND(value));
    }
    break;
  case FFF_INT:
    {
      int* buf = (int*)thisone->data;
      buf += pos;
      *buf = (int)(FFF_ROUND(value));
    }
    break;
  case FFF_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      buf += pos;
      *buf = (unsigned long int)(FFF_ROUND(value));
    }
    break;
  case FFF_LONG:
    {
      long int* buf = (long int*)thisone->data;
      buf += pos;
      *buf = (long int)(FFF_ROUND(value));
    }
    break;
  case FFF_FLOAT:
    {
      float* buf = (float*)thisone->data;
      buf += pos; 
      *buf = (float)value;
    }
    break;
  case FFF_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      buf += pos; 
      *buf = (double)value;
    }
    break;
  default: 
    fff_warning( proc, "Unrecognized image type.\n" );
    break; 
  }
  
  return;
} 


fff_array fff_array_get_block( const fff_array* thisone, 
			       size_t x0, size_t x1, size_t fX,
			       size_t y0, size_t y1, size_t fY,
			       size_t z0, size_t z1, size_t fZ,
			       size_t t0, size_t t1, size_t fT )
{
  char* proc = "fff_array_get_block"; 
  fff_array res; 
  fff_array_ndims ndims = FFF_ARRAY_4D; 
  size_t offX = thisone->offsetX; 
  size_t offY = thisone->offsetY; 
  size_t offZ = thisone->offsetZ; 
  size_t offT = thisone->offsetT; 

  /* Warning: coordinate range is not checked! */ 
  
  if ( t1 == t0 ) {
    ndims = FFF_ARRAY_3D; 
    if ( z1 == z0 ) {
      ndims = FFF_ARRAY_2D;
      if ( y1 == y0 ) 
	ndims = FFF_ARRAY_1D;
    }
  }
  res.ndims = ndims; 

  res.dimX = (x1-x0)/fX + 1; 
  res.dimY = (y1-y0)/fY + 1; 
  res.dimZ = (z1-z0)/fZ + 1;
  res.dimT = (t1-t0)/fT + 1; 

  res.offsetX = offX * fX; 
  res.offsetY = offY * fY; 
  res.offsetZ = offZ * fZ; 
  res.offsetT = offT * fT; 

  res.datatype = thisone->datatype; 

  /* Data pointer */ 
  res.owner = 0; 
  res.data = NULL; 
  switch(thisone->datatype) {
    
  case FFF_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case FFF_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case FFF_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case FFF_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf; 
    }
    break;
  case FFF_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case FFF_INT:
    {
      int* buf = (int*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf; 
    }
    break;
  case FFF_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case FFF_LONG:
    {
      long int* buf = (long int*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf; 
    }
    break;
  case FFF_FLOAT:
    {
      float* buf = (float*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case FFF_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf; 
    }
    break;
  default: 
    fff_warning( proc, "Unrecognized image type.\n" );
    break; 
  }
  
  return res; 
}



int fff_array_has_integer_data ( const fff_array* thisone )
{
  int ok; 

  switch ( thisone->datatype ) {

  default:
    ok = 0; 
    break; 

  case FFF_UCHAR:
  case FFF_SCHAR:
  case FFF_USHORT:
  case FFF_SSHORT:
  case FFF_UINT:
  case FFF_INT:
  case FFF_ULONG:
  case FFF_LONG:
    ok = 1; 
    break;
  
  }

  return ok; 
}

void fff_array_extrema ( double* min, double* max, const fff_array* thisone ) 
{
  double val; 
  fff_array_iterator iter = fff_array_iterator_init( thisone ); 
  size_t i;
  
  /* Initialization */ 
  *min = 0.0; 
  *max = 0.0; 

  i = 0; 
  while ( iter.uncomplete ) {
    val = fff_array_get_from_index( thisone, i ); 
    if ( val < *min )
      *min = val; 
    else if ( val > *max ) 
      *max = val;
    i += fff_array_iterator_update(&iter); 
  }

  return; 
}


/*
  Applies an affine correction to the input array so that:

  s0 --> r0 
  s1 --> r1

*/

void fff_array_compress( fff_array* ares, const fff_array* asrc, 
			 double r0, double s0, 
			 double r1, double s1 )
{
  char* proc = "fff_array_compress"; 

  /* Check dimensions */ 
  if ( (ares->dimX != asrc->dimX) ||
       (ares->dimY != asrc->dimY) ||
       (ares->dimZ != asrc->dimZ) ||
       (ares->dimT != asrc->dimT) ) {
    fff_warning( proc, "Arrays have different sizes.\n" );  
    return; 
  }
  
  fff_array_iterator itSrc = fff_array_iterator_init( asrc ); 
  fff_array_iterator itRes = fff_array_iterator_init( ares ); 
  size_t i, j;
  double a, b, valSrc; 

  a = (r1-r0) / (s1-s0); 
  b = r0 - a*s0; 

  i = 0; 
  j = 0; 
  while ( itSrc.uncomplete ) {
    
    valSrc = fff_array_get_from_index( asrc, i ); 
    fff_array_set_from_index( ares, j, a*valSrc+b ); 
    i += fff_array_iterator_update( &itSrc ); 
    j += fff_array_iterator_update( &itRes ); 

  }

  return;
}






fff_array_iterator fff_array_iterator_init(const fff_array* im)
{
  fff_array_iterator thisone;
  size_t pX, pY, pZ, pT;  
  thisone.x = 0; 
  thisone.y = 0; 
  thisone.z = 0; 
  thisone.t = 0; 
  thisone.ddimX = im->dimX - 1;
  thisone.ddimY = im->dimY - 1; 
  thisone.ddimZ = im->dimZ - 1; 
  thisone.ddimT = im->dimT - 1;
  pX = thisone.ddimX * im->offsetX; 
  pY = thisone.ddimY * im->offsetY; 
  pZ = thisone.ddimZ * im->offsetZ; 
  pT = thisone.ddimT * im->offsetT; 
  thisone.incT = im->offsetT; 
  thisone.incZ = im->offsetZ - pT; 
  thisone.incY = im->offsetY - pZ - pT; 
  thisone.incX = im->offsetX - pY - pZ - pT; 
  thisone.axis = -1; 
  thisone.uncomplete = 1; 
  return thisone; 
}

fff_array_iterator fff_array_iterator_init_skip_axis(const fff_array* im, int axis)
{
  size_t pX, pY, pZ, pT;  
  fff_array_iterator thisone = fff_array_iterator_init(im);  
  if ( (axis>=0) && (axis<4) )
    thisone.axis = axis; 
  pX = thisone.ddimX * im->offsetX; 
  pY = thisone.ddimY * im->offsetY; 
  pZ = thisone.ddimZ * im->offsetZ; 
  pT = thisone.ddimT * im->offsetT; 
  if (axis == 3)
    pT = 0; 
  else if (axis == 2)
    pZ = 0; 
  else if (axis == 1) 
    pY = 0; 
  else /* axis == 0 */ 
    pX = 0; 
  thisone.incT = im->offsetT; 
  thisone.incZ = im->offsetZ - pT; 
  thisone.incY = im->offsetY - pZ - pT; 
  thisone.incX = im->offsetX - pY - pZ - pT; 
  return thisone;
}



/* Return the index offset to access next element */ 
size_t fff_array_iterator_update( fff_array_iterator* thisone )
{
  int axis = thisone->axis; 

  if ( (thisone->t < thisone->ddimT) && (axis!=3) ) {
    thisone->t ++; 
    return thisone->incT; 
  }

  if ( (thisone->z < thisone->ddimZ) && (axis!=2) ) {
    thisone->t = 0; 
    thisone->z ++;
    return thisone->incZ; 
  }  

  if ( (thisone->y < thisone->ddimY) && (axis!=1) ) {
    thisone->t = 0; 
    thisone->z = 0; 
    thisone->y ++; 
    return thisone->incY;
  } 

 if ( (thisone->x < thisone->ddimX) && (axis!=0) ) {
    thisone->t = 0; 
    thisone->z = 0; 
    thisone->y = 0; 
    thisone->x ++; 
    return thisone->incX; 
  } 

  /* We are here because one can't move anymore */ 
  thisone->t = 0; 
  thisone->z = 0; 
  thisone->y = 0; 
  thisone->x = 0; 
  thisone->uncomplete = 0; 
  return 0; 
}



/* Image must be in DOUBLE format */ 
void fff_array_iterate_vector_function( fff_array* im, int axis, void(*func)(fff_vector*, void*), void* par )
{
  char* proc = "fff_array_iterate_vector_function"; 
  fff_array_iterator iter; 
  size_t i, size, stride; 
  double* im_data = (double*)im->data; 
  fff_vector x; 

  if ( im->datatype != FFF_DOUBLE ) {
    fff_warning( proc, "Image type must be double.\n" );
    return; 
  }
  if ( (axis>3) || (axis<0) ) {
    fff_warning( proc, "Invalid axis.\n" );
    return; 
 }
   
  size = fff_array_dim( im, axis );
  stride = fff_array_offset( im, axis );

  x.size = size; 
  x.stride = stride; 
  x.owner = 0; 

  iter = fff_array_iterator_init_skip_axis( im, axis );
  i = 0; 
  while ( iter.uncomplete ) {
    x.data = im_data + i;
    (*func)( &x, par );  
    i += fff_array_iterator_update(&iter); 
  }
  
  return; 
}




/* 
  Convert image values to [0,clamp-1]; typically clamp = 256.
  Possibly modify the dynamic range if the input value is
  overestimated.  For instance, the reconstructed MRI signal is
  generally encoded in 12 bits (values ranging from 0 to
  4095). Therefore, this operation may result in a loss of
  information.
*/

void fff_array_clamp( fff_array* ares, const fff_array* asrc, double th, int* clamp )
{
  char* proc = "fff_array_clamp"; 
  double imin, imax, tth; 
  int dmax = *clamp - 1;
 
  /* Compute input image min and max */
  fff_array_extrema( &imin, &imax, asrc ); 

  /* Make sure the threshold is not below the min intensity */ 
  tth = FFF_MAX(th, imin);

  /* Test */ 
  if (tth>imax) {
    fff_warning( proc, "Inconsistent threshold, ignored.\n" );
    tth = imin; 
  }    

  /* If the image dynamic is small, no need for compression: just
     downshift image values and re-estimate the dynamic range (hence
     imax is translated to imax-tth casted to SSHORT) */
  if ( ( fff_array_has_integer_data(asrc) ) && ( (imax-tth)<=dmax ) ) {
    fff_array_compress( ares, asrc, 0, tth, 1, tth+1 ); 
    *clamp = (int)(imax-tth) + 1;
  }
 
  /* Otherwise, compress after downshifting image values (values equal
     to the threshold are reset to zero) */ 
  else 
    fff_array_compress( ares, asrc, 0, tth, dmax, imax ); 

  return;
}
