#include "fff.h"
#include "fff_lapack.h"

#include <f2c.h>

#include <errno.h>

/*
dgetrf : LU decomp
dpotrf: Cholesky decomp
dgesdd: SVD decomp
dgeqrf: QR decomp
*/

#define CHECK_SQUARE(fn,A)						\
  if ( (A->size1) != (A->size2) )					\
    fff_error(fn, "Not a square matrix", EDOM)

#define LAPACK_UPLO(Uplo) ( (Uplo)==(CblasUpper) ? "U" : "L" )

extern int dgetrf_(integer *m, integer *n, doublereal *a, integer *
		   lda, integer *ipiv, integer *info);
extern int dpotrf_(char *uplo, integer *n, doublereal *a, integer *
		   lda, integer *info); 
extern int dgesdd_(char *jobz, integer *m, integer *n, doublereal *
		   a, integer *lda, doublereal *s, doublereal *u, integer *ldu,
		   doublereal *vt, integer *ldvt, doublereal *work, integer *lwork,
		   integer *iwork, integer *info);
extern int dgeqrf_(integer *m, integer *n, doublereal *a, integer *
		   lda, doublereal *tau, doublereal *work, integer *lwork, integer *info);


/* Cholesky decomposition */ 
/*** Aux needs be square with the same size as A ***/
int fff_lapack_dpotrf( CBLAS_UPLO_t Uplo, fff_matrix* A, fff_matrix* Aux )
{
  char* uplo = LAPACK_UPLO(Uplo);
  int info; 
  int n = (int)A->size1; /* Assumed squared */ 
  int lda = (int)Aux->tda; 
  
  CHECK_SQUARE("fff_lapack_dpotrf", A); 
  
  fff_matrix_transpose( Aux, A ); 
  dpotrf_(uplo, &n, Aux->data, &lda, &info); 
  fff_matrix_transpose( A, Aux ); 
  
  return info; 
}

/* LU decomposition */ 
/*** Aux needs be m x n with m=A->size2 and n=A->size1 ***/
/*** ipiv needs be 1d contiguous in int with size min(m,n) ***/
int fff_lapack_dgetrf( fff_matrix* A, fff_array* ipiv, fff_matrix* Aux )
{
  int info; 
  int m = (int)A->size1; 
  int n = (int)A->size2; 
  int lda = (int)Aux->tda; 
  
  if ( (ipiv->ndims != 1) || 
       (ipiv->datatype != FFF_INT) ||
       (ipiv->dimX != FFF_MIN(m,n)) ||
       (ipiv->offsetX != 1) ) 
    fff_error("fff_lapack_dgetrf", "Invalid array: Ipiv", EDOM); 

  fff_matrix_transpose( Aux, A );
  dgetrf_(&m, &n, Aux->data, &lda, (int*)ipiv->data, &info);
  fff_matrix_transpose( A, Aux ); 
  
  return info; 
}

/* QR decomposition */ 
/*** Aux needs be m x n with m=A->size2 and n=A->size1 ***/
/*** tau needs be contiguous with size min(m,n) ***/
/*** work needs be contiguous with size >= n ***/
int fff_lapack_dgeqrf( fff_matrix* A, fff_vector* tau, fff_vector* work, fff_matrix* Aux )
{
  int info; 
  int m = (int)A->size1;
  int n = (int)A->size2;
  int lda = (int)Aux->tda; 
  int lwork = (int)work->size; 

  if ( (tau->size != FFF_MIN(m,n)) ||
       (tau->stride != 1) )
    fff_error("fff_lapack_dgeqrf", "Invalid vector: tau", EDOM); 
    
  /* Resets lwork to -1 if the input work vector is too small (in
     which case work only needs be of size >= 1) */ 
  if ( lwork < n ) 
    lwork = -1;
  else 
    if ( work->stride != 1 )  
      fff_error("fff_lapack_dgeqrf", "Invalid vector: work", EDOM); 

  fff_matrix_transpose( Aux, A );
  dgeqrf_(&m, &n, Aux->data, &lda, tau->data, work->data, &lwork, &info); 
  fff_matrix_transpose( A, Aux ); 
  
  return info; 
}


/* SVD decomposition */ 
/*** Aux needs be square with size max(m=A->size2, n=A->size1) ***/
/*** s needs be contiguous with size min(m,n) ***/ 
/*** U needs be m x m ***/
/*** Vt needs be n x n ***/
/*** work needs be contiguous, with size lwork such that
dmin = min(M,N)
dmax = max(M,N)

lwork >=  3*dmin**2 + max(dmax,4*dmin**2+4*dmin)

 ***/
/*** iwork needs be 1d contiguous in int with size 8*min(m,n) ***/
int fff_lapack_dgesdd( fff_matrix* A, fff_vector* s, fff_matrix* U, fff_matrix* Vt, 
		       fff_vector* work, fff_array* iwork, fff_matrix* Aux )
{
  int info; 
  int m = (int)A->size1; 
  int n = (int)A->size2; 
  int dmin = FFF_MIN(m,n); 
  int dmax = FFF_MAX(m,n); 
  int a1 = FFF_SQR(dmin);
  int a2 = 4*(a1+dmin); 
  int lwork_min = 3*a1 + FFF_MAX(dmax, a2); 
  int lda = (int)Aux->tda; 
  int ldu = (int)U->tda; 
  int ldvt = (int)Vt->tda;
  int lwork = work->size;  

  fff_matrix Aux_mm, Aux_nn; 
  
  CHECK_SQUARE("fff_lapack_dgesdd", U);
  CHECK_SQUARE("fff_lapack_dgesdd", Vt);
  CHECK_SQUARE("fff_lapack_dgesdd", Aux);
  if ( U->size1 != m) 
    fff_error("fff_lapack_dgesdd", "Invalid size for U", EDOM); 
  if ( Vt->size1 != n) 
    fff_error("fff_lapack_dgesdd", "Invalid size for Vt", EDOM);
  if ( Aux->size1 != dmax) 
    fff_error("fff_lapack_dgesdd", "Invalid size for Aux", EDOM);  
  if ( (s->size != dmin) ||
       (s->stride != 1) )
    fff_error("fff_lapack_dgesdd", "Invalid vector: s", EDOM); 
  if ( (iwork->ndims != 1) || 
       (iwork->datatype != FFF_INT) ||
       (iwork->dimX != 8*dmin) ||
       (iwork->offsetX != 1 ) ) 
    fff_error("fff_lapack_dgesdd", "Invalid array: Iwork", EDOM);
  
  /* Resets lwork to -1 if the input work vector is too small (in
     which case work only needs be of size >= 1) */ 
  if ( lwork < lwork_min ) 
    lwork = -1;
  else 
    if ( work->stride != 1 )  
      fff_error("fff_lapack_dgesdd", "Invalid vector: work", EDOM); 
  
  /* 
     Perform the svd on A**t: 
     A**t = U* S* Vt* 
     => A = V* S* Ut*
     => U = V*, V = U*, s = s* 
     so we just need to swap m <-> n, and U <-> Vt in the input line
  */
  dgesdd_("A", &n, &m, A->data, &lda, 
	  s->data, Vt->data, &ldvt, U->data, &ldu, 
	  work->data, &lwork, (int*)iwork->data, &info);

  /* At this point, both U and V are in Fortran order, so we need to
     transpose */
  Aux_mm = fff_matrix_block( Aux, 0, m, 0, m );
  fff_matrix_transpose(&Aux_mm, U); 
  fff_matrix_memcpy(U, &Aux_mm); 
  Aux_nn = fff_matrix_block( Aux, 0, n, 0, n ); 
  fff_matrix_transpose(&Aux_nn, Vt); 
  fff_matrix_memcpy(Vt, &Aux_nn); 

  return info; 
}

