#include <fff.h>
#include <fff_vector.h>
#include <fff_array.h>

#include <mex.h>

/*!
  \file fffMat.h
  \brief Matlab interface to \a fff
  \author Alexis Roche
  \date 2006-2008

  Any conversion from an \c mxArray to one of the supported C
  structures returns a view from the input \c mxArray, hence it NEVER
  involves copying array elements. In a MEX function, you can use
  those functions to convert both input and output mxArrays (outputs
  need be allocated beforehand using an appropriate mxCreate<>
  function).
  
  Owing to this full compatibility, it would be pointless to provide
  conversion functions from C structs to mxArrays. By the way, such
  functions would necessitate copies because the \c mxArray API does
  not seem to provide means to create an \c mxArray from an already
  allocated buffer (contrary to Numpy).
  
*/ 

/*! 
  \brief Convert an \c mxArray to a \c fff_vector 
  \param x input array 

  This is a \c fff_vector constructor compatible with \c fff_vector_delete
*/ 
extern fff_vector* fff_vector_from_mxArray( const mxArray *x );

/*! 
  \brief Convert a Matlab matrix into the list of its column vectors 
  \param x input matrix as an \c mxArray 
  \param nc returned number of fff_vectors

  Matlab matrices are stored in column-major order (Fortran
  convention) as opposed to the row-major order underlying the
  GSL. Converting a Matlab matrix into a list of vectors avoids
  copying array elements.
*/  
extern fff_vector** fff_vector_list_from_mxArray_new( const mxArray *x, int* nc );
/*! 
  \brief Delete a list of fff_vectors created by \c fff_vector_list_from_mxArray_new
  \param x list of vectors
  \param nc number of vectors
  
*/  
extern void fff_vector_list_from_mxArray_delete( fff_vector **x, int nc );

/*! 
  \brief Convert an \c mxArray to an \c fff_array
  \param x input array 

  This is an \c fff_array constructor compatible with \c fff_array_delete
*/ 
extern fff_array* fff_array_from_mxArray( const mxArray *x ); 

/*! 
 \brief Checker whether an \c mxArray is in \c long format 
 \param x input array
 \param mxLong returned Matlab class matching the \c long format
*/ 
extern int mxArray_is_long( const mxArray*x, mxClassID* mxLong ); 



