#include "fffMat.h"

#include <fff_graphlib.h>
#include <fff_field.h>

#include <string.h>

/* 

SYNTAX: [size, label, depth] = fffMat_clusters( XYZ, T ); 

Warning: the input array XYZ must be in C-compatible long int format
(most often, 32 bits or 64 bits).

The returned arrays (size, label, depth) are all C-compatible long int
arrays.

*/ 

void mexFunction( int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray*prhs[] )
{ 
  mxClassID mxLong;
  gsl_matrix_long XYZ; 
  gsl_vector* T;
  fff_graph* G = NULL; 
  long *label, *bufL, *size; 
  gsl_vector_long depth; 
  int nvox, edges, clusters, maxima; 
  int i, j; 

  /* Check for proper number of arguments. */
  if(nrhs < 2) {
    mexErrMsgTxt("Two input arguments required.");
  } 
  else if(nlhs > 3) {
    mexErrMsgTxt("Too many output arguments");
  }
  
  /* Input */ 
  nvox = mxGetN( prhs[0] ); 

  /* Consistency check */ 
  if ( ( mxGetM( prhs[0] )!= 3 ) ||
       ( mxGetN( prhs[1] )!= nvox ) )
    mexErrMsgTxt("Inconsistent input arrays.");
  
  /* Check input type */ 
  if ( ! mxArray_is_long( prhs[0], &mxLong ) )
    mexErrMsgTxt("Input array XYZ is not in proper int format.");

  /* XYZ matrix */ 
  XYZ.size1 = nvox; 
  XYZ.size2 = 3; 
  XYZ.tda = 3; 
  XYZ.data = (long*) mxGetPr( prhs[0] ); 
  XYZ.owner = 0; 
  XYZ.block = NULL; 
  
  /* Field values */ 
  T = gsl_vector_from_mxArray( prhs[1] ); 
  
  /* Output */ 
  plhs[1] = mxCreateNumericMatrix( nvox, 1, mxLong, mxREAL );
  plhs[2] = mxCreateNumericMatrix( nvox, 1, mxLong, mxREAL );

  /* Make a graph out of the input coordinates */
  edges = fff_graph_grid( &G, &XYZ, 18 );
  
  /* Assign cluster label to each voxel */ 
  label = (long*) mxGetPr( plhs[1] );
  clusters = fff_graph_cc_label( label, G );

  /* Compute the cluster sizes */ 
  plhs[0] = mxCreateNumericMatrix( clusters, 1, mxLong, mxREAL );
  size = (long*) mxGetPr( plhs[0] );
  bufL = label; 
  for( j=0; j<nvox; j++, bufL++ ) {
    i = *bufL; 
    size[i] += 1.0; 
    *bufL = i+1; /* To return indices starting from 1... */ 
  }

  /* Compute depth for each voxel */
  depth.size = (size_t) nvox; 
  depth.stride = 1; 
  depth.data = (long*) mxGetPr( plhs[2] ); 
  depth.owner = 0; 
  depth.block = NULL; 
  maxima = fff_field_maxima( &depth, G, T ); 

  /* Free memory */ 
  fff_graph_delete( G ); 
  gsl_vector_free( T ); 

  return; 

}


