#include "fffMat.h"

#include <fff_onesample_stat.h>
#include <gsl/gsl_rng.h>

#include <string.h>
#include <time.h>

#define DEFAULT_FLAG FFF_ONESAMPLE_STUDENT
#define DEFAULT_BASE 0.0
#define DEFAULT_NSIMU 1
#define DEFAULT_MAGIC 0
#define DEFAULT_NITER 5

#define MFX_OFFSET 10 

/* 
   
T = fffMat_onesample_stat( E, V, flagStat, base, nsimu, magic, niter ); 

magic < 0    ==> randomize permutations
magic >= 0   ==> produce deterministic permutations in the range [magic, magic+nsimu[

*/ 

static void _set_simulation_parameters( int* flag_permute, int* flag_use_rng, 
					int* simu0, int* nsimu, 
					size_t* nsimu_max, int n, 
					long int magic )
{
  if ( *nsimu < 1 )
    *nsimu = 1;     
  *flag_permute = ( (*nsimu > 1) || (magic!=0) ); 
  *flag_use_rng = ( magic < 0 );
  
  *simu0 = 0; 
  if ( magic == 0 )
    *simu0 = 1; 

  *nsimu_max = 1 << n; 
  if ( *flag_permute ) 
    if ( ( (size_t)(*nsimu) > *nsimu_max ) && 
	 ( !(*flag_use_rng) ) )
      *nsimu = *nsimu_max;
  
  return; 
}

void mexFunction( int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray*prhs[] )
{ 
  /* Variables */ 
  int flag = DEFAULT_FLAG;
  double base = DEFAULT_BASE; 
  int nsimu = DEFAULT_NSIMU;
  long int magic = DEFAULT_MAGIC; 
  int niter = DEFAULT_NITER;
  gsl_vector **E, **V, **T;
  int vox, nvox, nV, simu, simu0;
  fff_onesample_stat* Stat; 
  fff_onesample_stat_mfx* StatMFX; 
  size_t nsimu_max; 
  int flag_permute, flag_use_rng, flag_mfx = 0;
  long int magic_aux; 
  gsl_rng* rng = NULL;
  double *aux; 

  /* Check for proper number of arguments. */
  if ( nrhs < 2 ) {
    mexErrMsgTxt("Two input arguments required.");
  } 
  else if ( nlhs > 50 ) {
    mexErrMsgTxt("Too many output arguments");
  }

  /* Get list of effect vectors */
  E = gsl_vector_list_from_mxArray_new( prhs[0], &nvox );
  
  /* Get list of associated variances */
  V = gsl_vector_list_from_mxArray_new( prhs[1], &nV );
  
  /* Test consistency between E and V */
  if ( (nV != 0) && (nV != nvox) ) {
    gsl_vector_list_from_mxArray_delete( E, nvox );
    gsl_vector_list_from_mxArray_delete( V, nV );
    mexErrMsgTxt("Incompatible effect and variance arrays.");
  }

  /* Statistic choice */
  if ( nrhs > 2 ) {
    aux = mxGetPr( prhs[2] );
    flag = (int)(*aux);
    if ( flag >= MFX_OFFSET ) 
      flag_mfx = 1; 
  }
  
  /* Baseline */
  if ( nrhs > 3 ) {
    aux = mxGetPr( prhs[3] );
    base = (double)(*aux);
  }

  /* Number of simulations */
  if ( nrhs > 4 ) {
    aux = mxGetPr( prhs[4] );
    nsimu = (int)(*aux);
  }

  /* Magic number */ 
  if ( nrhs > 5 ) {
    aux = mxGetPr( prhs[5] );
    magic = (long int)(*aux);
  }

  /* Number of EM iterations (MFX stats) */ 
  if ( nrhs > 6 ) {
    aux = mxGetPr( prhs[6] );
    niter = (int)(*aux);
  }

  
  /* Simulation parameters */ 
  _set_simulation_parameters( &flag_permute, &flag_use_rng, 
			      &simu0, &nsimu, &nsimu_max, 
			      (int)E[0]->size, magic ); 

  /* Output */ 
  plhs[0] = mxCreateDoubleMatrix( nsimu, nvox, mxREAL );
  
  /* Local structures */
  T = gsl_vector_list_from_mxArray_new( plhs[0], &nvox );
  if ( ! flag_mfx )
    Stat = fff_onesample_stat_new( (size_t)E[0]->size, (fff_onesample_stat_flag)flag, base );
  else {
    StatMFX = fff_onesample_stat_mfx_new( (size_t)E[0]->size, (fff_onesample_stat_mfx_flag)(flag-MFX_OFFSET), base );
    fff_onesample_stat_mfx_set_niter( StatMFX, (unsigned int)niter );
  }
  if ( flag_use_rng ) {
    gsl_rng_env_setup();
    rng = gsl_rng_alloc( gsl_rng_default );
  }

  /* Compute stat for the original dataset */ 
  if ( simu0 == 1 ) 
    if ( ! flag_mfx ) 
      for ( vox=0; vox<nvox; vox++ ) {
	fff_onesample_stat_eval( Stat, E[vox] );
	T[vox]->data[0] = Stat->stat; 
      }
    else 
      for ( vox=0; vox<nvox; vox++ ) {
	fff_onesample_stat_mfx_eval( StatMFX, E[vox], V[vox] );
	T[vox]->data[0] = StatMFX->stat; 
      }

  /* Simulations */ 
  for ( simu=simu0; simu<nsimu; simu++ ) {

    /* Determine permutation (magic number) */ 
    if ( flag_use_rng ) 
      magic_aux = gsl_rng_uniform_int( rng, nsimu_max );
    else 
      magic_aux = (size_t)(simu + magic); 
    
    /* Loop over voxels */
    if ( ! flag_mfx ) {
      Stat->magic = magic_aux; 
      for ( vox=0; vox<nvox; vox++ ) {
	fff_onesample_stat_permute_and_eval( Stat, E[vox] );   
	T[vox]->data[simu] = Stat->stat;
      }
    }
    else {
      StatMFX->magic = magic_aux; 
      for ( vox=0; vox<nvox; vox++ ) {
	fff_onesample_stat_mfx_permute_and_eval( StatMFX, E[vox], V[vox] );   
	T[vox]->data[simu] = StatMFX->stat;
      }
    }
  
  } /* End loop over simulations */ 

  /* Free memory */ 
  gsl_vector_list_from_mxArray_delete( E, nvox );
  gsl_vector_list_from_mxArray_delete( V, nV );
  gsl_vector_list_from_mxArray_delete( T, nvox );
  if ( ! flag_mfx ) 
    fff_onesample_stat_delete( Stat ); 
  else 
    fff_onesample_stat_mfx_delete( StatMFX ); 
  if ( flag_use_rng ) 
    gsl_rng_free( rng ); 

  return; 
}

