function setup( flagSTATIC, lapack_prefix, fff_prefix ) 
%
% COMPILE THE FFFMAT MEX-FILES REQUIRED BY DISTANCE TOOLBOX
%_______________________________________________________________________
%
% FORMAT  setup( flagSTATIC, lapack_prefix, fff_prefix )
%
% *** INPUTS ***
%  - 'flagSTATIC' (INTEGER 0 or 1, not required, default = 1)
% --> Flag for specifying type of linking with LAPACK and FFF core
%     libraries:
%  *  flagSTATIC = 1 --> Force static linking
%  *  flagSTATIC = 0 --> Perform dynamic linking
%
%  - 'lapack_prefix' ([1xn] CHARACTER ARRAY CONTAINING PATHNAME TO THE LAPACK
%                INSTALLATION DIRECTORY, not required, default = '')
% --> If 'lapack_prefix' is unspecified, an attempt to autodetect the
%     pathname to the LAPACK installation directory is performed.
% --> Be careful that the MEX-files compilation will be aborted if the
%     LAPACK installation directory is not found.
%
%  - 'fff_prefix' ([1xp] CHARACTER ARRAY CONTAINING PATHNAME TO THE FFF
%                CORE LIBRARY, not required, default = '')
% --> If 'fff_prefix' is unspecified, an attempt to autodetect the
%     pathname to an existing FFF core library is performed. Upon
%     failure, the core FFF library will be compiled on the fly and
%     linked statically. 
% --> Be careful that the MEX-files compilation will be aborted if an
%     existing FFF core library cannot be linked.
%_______________________________________________________________________
%
% This function allows to compile all the fffMat MEX-files required by
% DISTANCE Toolbox using either static or dynamic linking with the LAPACK
% and FFF core libraries. It can only run from the directory where it is
% located in the FFF tree.
%
% The compilation will generated three MEX-files (extension depending on
% type of operating system):
% - fffMat_clusters.*
% - fffMat_onesample_stat.*
% - fffMat_quantile.*
% After compilation, these MEX-files must be moved into the
% installation directory of DISTANCE Toolbox in order to enable
% operation of group analysis module.
%
% If you are encountered any problems during the compilation process,
% please contact us at distance-toolbox@googlegroups.com.
%_______________________________________________________________________
% @ A. Roche & S. Mériaux - CEA/DSV/I2BM/NEUROSPIN/LNAO

%----------------------------------------------------------------------%
% @ Copyright (C) 1997-2008 CEA/INRIA/INSERM                           %
% @ For bugs report, remarks, comments,... please contact directly the %
% @ authors at: distance-toolbox@googlegroups.com                      %
% @ WARNING: THIS SOFTWARE IS DISTRIBUTED FREE WITHOUT ANY WARRANTY    %
%----------------------------------------------------------------------%

%=======================================================================
% DISTANCE FUNCTION NAME AND VERSION
%=======================================================================
dist_fcn = 'setup' ;
dist_vers = 'v_3.XX' ;

%=======================================================================
% Define input arguments if unspecified
%=======================================================================
if ( nargin < 1 ), flagSTATIC = 1 ; end
if ( isempty( flagSTATIC ) == 1 ), flagSTATIC = 1 ; end
if ( nargin < 2 ), lapack_prefix = '' ; end
if ( isempty( lapack_prefix ) == 1 ), lapack_prefix = '' ; end
if ( nargin < 3 ), fff_prefix = '' ; end
if ( isempty( fff_prefix ) == 1 ), fff_prefix = '' ; end

%=======================================================================
% START "setup.m" FUNCTION:
% COMPILE FFFMAT MEX-FILES
%=======================================================================

%-----------------------------------------------------------------------
% START FUNCTION MESSAGE
%-----------------------------------------------------------------------
fprintf( 1, '\n *** FUNCTION "%s.m" (%s)\n', dist_fcn, dist_vers ) ;

%-----------------------------------------------------------------------
% Load specific compilation parameters (depending on type of operating
% system)
%-----------------------------------------------------------------------
fprintf( 1, ' +++ Load compilation parameters...\n' ) ;
% Define compilation parameters
install_path = pwd ;
mex_cmd = 'mex ' ;
mex_copt = '-c -O' ;
mex_mopt = '-O' ;
static_ext = '.a' ;
includelapack_path = 'include' ;
liblapack_path = 'lib' ;
liblapack = 'liblapack' ;
ffflite_path = fileparts( install_path ) ;
includefff_path = 'core' ;
buildfff_path = 'build' ;
aux_file = 'temp.' ;
libfffcore_path = 'libfffcore' ;
% Define additionnal compilation parameters depending on OS type
if ( ispc )
   dll_ext = '.dll.a' ;
   obj_ext = '.obj' ;
   lapack_prefix_def = { 'C:\cygwin\usr\local', 'C:\msys\1.0\local' } ;
else
   dll_ext = '.so' ;
   obj_ext = '.o' ;
   lapack_prefix_def = { '/usr/local' } ;
end

%-----------------------------------------------------------------------
% Modify the 'CLIBS' definition in MEX's options file: remove the link
% with stdc++ library
%
% NB. This link is responsible for execution errors of MEX-files when
%     the gcc versions used by Matlab and by operating system are not
%     compatible. 
%-----------------------------------------------------------------------
% Determine additionnal argument for 'mex.m' function which removes the
% link with stdc++ library
mex_addopt = sf_modify_clibs( fullfile( install_path, 'fffMat.c' ) ) ;

%-----------------------------------------------------------------------
% Try to autodetect the pathname to the LAPACK installation directory if
% 'lapack_prefix' is unspecified
%-----------------------------------------------------------------------
fprintf( 1, [ ' +++ Load pathname to the LAPACK installation', ...
   ' directory...\n' ] ) ;
status = 1 ;
% Try to autodetect the pathname to the LAPACK installation directory if
% 'lapack_prefix' is unspecified
% Check existence of the specified LAPACK installation directory
[lapack_prefix, status] = sf_checkdir( lapack_prefix ) ;
if ( status > 0 )
   fprintf( 1, [ ' WARNING: wrong specified LAPACK installation', ...
         ' directory\n' ] ) ;
end
% Valid LAPACK installation directory found
if ( status < 1 )
   fprintf( 1, ' --> LAPACK installation directory:\n' ) ;
   fprintf( 1, '     ''%s''\n', lapack_prefix ) ;
else
   % Try to go to standard LAPACK installation directory(ies)
   i = 1 ;
   while ( status > 0 && i <= length( lapack_prefix_def ) )
      fprintf( 1, ' --> Try standard LAPACK installation directory:\n' ) ;
      fprintf( 1, '     ''%s''\n', lapack_prefix_def{i} ) ;
      [lapack_prefix, status] = sf_checkdir( lapack_prefix_def{i} ) ;
      i = i + 1 ;
   end
   clear( 'i' ) ;
   % Cannot find any standard LAPACK installation directory
   if ( status > 0 )
      error( sprintf( [ 'Cannot find any standard LAPACK installation', ...
         ' directory...\n' ] ) ) ;
   end
end
% Clear unnecessary variables
clear( 'lapack_prefix_def', 'status' ) ;

%-----------------------------------------------------------------------
% Set LAPACK dependency  
%-----------------------------------------------------------------------
% Define 'include' directory
include_lapack = fullfile( lapack_prefix, includelapack_path ) ;
%
fprintf( 1, ' +++ Set LAPACK dependency...\n' ) ;
% Try to detect 'liblapack' library
[lib_lapack, ok] = sf_setlib( fullfile( lapack_prefix, liblapack_path, ...
   liblapack ), dll_ext, static_ext, flagSTATIC ) ;
% If LAPACK is not found, use the included LAPACK lite sources 
if ( ok < 1 ), 
    fprintf(1, 'No LAPACK library found on the system, building a lite LAPACK\n' ); 
    include_lapack = []; 
else, 
    fprintf( 1, 'Found ''%s''\n', lib_lapack ) ;
end,
clear( 'lapack_prefix', 'includelapack_path', 'liblapack_path', 'liblapack', 'ok' ) ;

% Alexis, 31/3/08. I am leaving autodetect fff for now but I think that feature is almost 
% useless in practice, if not dangerous. 

%-----------------------------------------------------------------------
% Set FFF dependency
%-----------------------------------------------------------------------
% Define expected directory for FFF core library
include_fff = fullfile( ffflite_path, includefff_path ) ;
%
fprintf( 1, ' +++ Set FFF dependency...\n' ) ;
% Try to autodetect a pre-compiled FFF shared library if 'fff_prefix' is
% unspecified
if ( isempty( fff_prefix ) == 1 )
   % Check content of 'build' sub-directory
   aux = dir( fullfile( ffflite_path, buildfff_path ) ) ;
   %
   [i, ok] = deal( 1, 0 ) ;
   while ( ok < 1 && i <= length( aux ) )
      i = i + 1 ;
      ok = strncmp( aux_file, aux(i).name, 4 ) ;
   end
   %
   if ( ok > 0 )
      [fff_prefix, ok] = sf_setlib( fullfile( ffflite_path, ...
         buildfff_path, aux(i).name, libfffcore_path ), dll_ext, ...
         static_ext, flagSTATIC ) ;
   end
   %
   if ( isequal( ok, 2 ) == 1 )
      fprintf( 1, ' --> Pre-compiled FFF library found in:\n' ) ;
      fprintf( 1, '     ''%s''\n', fff_prefix ) ;
   else
      fprintf( 1, ' WARNING: no pre-compiled FFF library found\n' ) ;
      fff_prefix = '' ;
   end
   %
   clear( 'aux', 'i', 'ok' ) ;
% Check the specified pathname to the FFF core library
else
   % Try to go to specified FFF core library
   [fff_prefix, status] = sf_checkdir( fff_prefix ) ;
   % Cannot find FFF core library
   if ( status > 0 )
      fprintf( 1, ' WARNING: wrong specified FFF core library\n' ) ;
   end
   %
   clear( 'status' ) ;
end
%
clear( 'flagSTATIC', 'ffflite_path', 'includefff_path', ...
   'buildfff_path', 'aux_file', 'libfffcore_path', 'static_ext', ...
   'dll_ext' ) ;

%-----------------------------------------------------------------------
% No linking against a pre-compiled FFF shared library found: compile
% FFF core as a collection of object files
%-----------------------------------------------------------------------
%
ObjFiles = {} ;
%
if ( isempty( fff_prefix ) == 1 )
   %
   fprintf( 1, ' --> Try to compile FFF core on the fly:\n' ) ;
   % Try to find *.c file(s)
   CoreFiles = dir( [ include_fff, filesep, 'fff*.c' ] ) ;
   if ( length( CoreFiles ) < 1 )
      error( sprintf( 'Cannot find any C-file to be compiled...\n' ) ) ;
   end
   %
   include_f2c = [include_fff, filesep, 'lapack_lite']; 
   if isempty( include_lapack )
     no_lapack = 1; 
     LapackFiles = dir([ include_fff, filesep, 'lapack_lite', filesep, '*.c' ] ) ;
     include_all = [ ' -I', include_fff, ' -I', include_f2c] ;
   else,
     no_lapack = 0;
     LapackFiles = [];
     include_all = [ ' -I', include_fff, ' -I', include_f2c, ' -I', include_lapack ] ;
   end, 
   lib_all = lib_lapack;
   % Compile C-file(s)
   if no_lapack, 
      ObjFiles = cell( 1, length( CoreFiles ) ) ;
   else, 
      ObjFiles = cell( 1, length( CoreFiles ) + length( LapackFiles) ) ;
   end,
   for i = 1:1:length( CoreFiles )
      % Generate C-filename to be compiled 
      Cfile = [ include_fff, filesep, CoreFiles(i).name ] ;
      fprintf( 1, ' ... Compiling C-file: ''%s''\n', Cfile ) ;
      % Try to compile C-file
      try
         eval( [ mex_cmd, mex_addopt, ' ', mex_copt, include_all, ...
            ' ', Cfile, ' ', lib_all ] ) ;
      catch
         error( sprintf( 'Compilation failed for ''%s''...\n', ...
            Cfile ) ) ;
      end
      % If compilation succeeds, store corresponding object file to be
      % used for MEX-files generation
      [niu, fileName] = fileparts( Cfile ) ;
      ObjFiles{i} = [ fileName, obj_ext, ' ' ] ;
   end
   if no_lapack,
      for i = 1:1:length( LapackFiles )
         % Generate C-filename to be compiled 
         Cfile = [ include_fff, filesep, 'lapack_lite', filesep, LapackFiles(i).name ] ;
         fprintf( 1, ' ... Compiling C-file: ''%s''\n', Cfile ) ;
         % Try to compile C-file
         try
            eval( [ mex_cmd, mex_addopt, ' ', mex_copt, include_all, ...
               ' ', Cfile, ' ', lib_all ] ) ;
         catch
            error( sprintf( 'Compilation failed for ''%s''...\n', ...
               Cfile ) ) ;
	 end
         % If compilation succeeds, store corresponding object file to be
         % used for MEX-files generation0
         [niu, fileName] = fileparts( Cfile ) ;
         ObjFiles{ length(CoreFiles) + i } = [ fileName, obj_ext, ' ' ] ;
      end
   end
   % Clear unnecessary variables
   clear( 'CoreFiles', 'lib_all', 'i', 'Cfile', 'niu', 'fileName' ) ;
end
%

%-----------------------------------------------------------------------
% Compile the FFF/Matlab wrapper (fffMat.c) and make an object file
%-----------------------------------------------------------------------
%
wrap_file = fullfile( install_path, 'fffMat.c' ) ;
fprintf( 1, ' --> Compile FFF/Matlab wrapper:\n' ) ;
fprintf( 1, '     ''%s''\n', wrap_file ) ;
%

include_all = [include_all, ' -I', install_path];
lib_all = [ cell2mat( ObjFiles ), fff_prefix, ' ', lib_lapack ];

% Try to compile C-file
try
   eval( [ mex_cmd, mex_addopt, ' ', mex_copt, include_all, ...
      ' ', wrap_file, ' ', lib_all ] ) ;
catch
   error( sprintf( 'Compilation failed for ''%s''...\n', wrap_file ) ) ;
end
% If compilation succeeds, store corresponding object file to be used
% for MEX-files generation0
[niu, fileName] = fileparts( wrap_file ) ;
ObjFiles{ length( ObjFiles ) + 1 } = [ fileName, obj_ext] ;
%
clear( 'mex_copt', 'wrap_file', 'obj_ext', ...
   'include_fff', 'include_lapack', 'niu', 'fileName' ) ;

%-----------------------------------------------------------------------
% Compile all MEX-files
%-----------------------------------------------------------------------
fprintf( 1, ' --> Compile MEX-files:\n' ) ;
%
MexFiles = dir( [ install_path, filesep, 'fffMat_*.c' ] ) ;
%
if ( length( MexFiles ) < 1 )
   error( sprintf( 'Cannot find any MEX-file to be compiled...\n' ) ) ;
end
%
lib_all = [ cell2mat( ObjFiles ), fff_prefix, ' ', lib_lapack ];
% Try to compile MEX-files
for i = 1:1:length( MexFiles )
   %
   Mfile = fullfile( install_path, MexFiles(i).name ) ;
   fprintf( 1, ' ... Compiling MEX-file: ''%s''\n', Mfile ) ;
   %
   try
      eval( [ mex_cmd, mex_addopt, ' ', mex_mopt, include_all, ...
         ' ', Mfile, ' ', lib_all ] ) ;
   catch
      error( sprintf( 'Compilation failed for ''%s''...\n', Mfile ) ) ;
   end
end
%
clear( 'install_path', 'fff_prefix', 'mex_cmd', 'mex_mopt', ...
   'lib_lapack', ...
   'mex_addopt', 'include_all', 'lib_all', 'MexFiles', 'i', 'Mfile' ) ;

%-----------------------------------------------------------------------
% Delete object files
%-----------------------------------------------------------------------
%
fprintf( 1, ' --> Delete object files:\n' ) ;
%
for i = 1:1:length( ObjFiles )
   %
   delObjFile = deblank( ObjFiles{i} ) ;
   delete( delObjFile ) ;
   %
   fprintf( 1, ' ... ''%s'' deleted\n', delObjFile ) ;
end
%
clear( 'ObjFiles', 'i', 'delObjFile' ) ;

%-----------------------------------------------------------------------
% END FUNCTION MESSAGE
%-----------------------------------------------------------------------
%
fprintf( 1, ' *** MEX-files compilation successful\n\n' ) ;
%
clear( 'dist_fcn', 'dist_vers' ) ;


%#######################################################################
%                                                                      %
%                     UTILITY  EMBEDDED  FUNCTION                      %
%                                                                      %
%#######################################################################

%#######################################################################
% I] UTILITY EMBEDDED FUNCTION: "sf_checkdir"
%
%#######################################################################
function  [valid_dir, status] = sf_checkdir( dir_check )


%=======================================================================
% Outputs initialisation
%=======================================================================
[valid_dir, status] = deal( '', 1 ) ;

%=======================================================================
% Check pathname to existing directory
%=======================================================================
% Store pathname to current directory
temp_dir = pwd ;
% Try to change to directory to be checked
try
   cd( dir_check ) ;
   valid_dir = pwd ;
   status = 0 ;
end
% Return to current directory
cd( temp_dir ) ;


%#######################################################################
% II] UTILITY EMBEDDED FUNCTION: "sf_modify_clibs"
%
%#######################################################################
function  mex_addopt = sf_modify_clibs( Cfile_test )


%=======================================================================
% Output initialisation
%=======================================================================
mex_addopt = '' ;

%=======================================================================
% Read arguments of 'mex.m' function
%=======================================================================
[niu, arg_str] = system( [ 'mex -v -n ', Cfile_test, ' ;' ] ) ;

%=======================================================================
% Try to find 'CLIBS' in list of arguments
%=======================================================================
% Try to find 'CLIBS' in list of arguments
indClibs = strfind( arg_str, 'CLIBS' ) ;
% No 'CLIBS' argument found
if ( isempty( indClibs ) == 1 ), return ; end

%=======================================================================
% Try to find first linefeed after 'CLIBS' in list of arguments
%=======================================================================
% Try to find first linefeed after 'CLIBS'
indLfeed = strfind( arg_str, sprintf( '\n' ) ) ;
% No linefeed found
if ( isempty( indLfeed ) == 1 ), return ; end
% No linefeed found after 'CLIBS'
% NB. +7 because expected syntax for 'CLIBS' argument is: CLIBS=*\n
if ( max( indLfeed ) < indClibs + 7 ), return ; end

%=======================================================================
% Get original value of 'CLIBS' argument
%=======================================================================
% Retrieve syntax: CLIBS=[clibs_arg]
indLfeed = indLfeed( find( indLfeed >= indClibs + 7 ) ) ;
Clibs = arg_str( indClibs : 1 : indLfeed(1) - 1 ) ;
% Retrieve original value of 'CLIBS' argument ([clibs_arg])
indClibs = strfind( Clibs, '=' ) ;
% No '=' character found
if ( isempty( indClibs ) == 1 ), return ; end
% Remove any blank character beginning or ending value of 'CLIBS'
% argument
Clibs = fliplr( deblank( fliplr( deblank( Clibs( indClibs + 1 : ...
   end ) ) ) ) ) ;
%
clear( 'niu', 'arg_str', 'indClibs', 'indLfeed' ) ;
%
fprintf( 1, ' ... Old CLIBS argument:\n' ) ;
fprintf( 1, '     CLIBS = %s\n', Clibs ) ;

%=======================================================================
% Remove link with stdc++ library in value of 'CLIBS' argument
%=======================================================================
%
Clibs = strrep( Clibs, ' -lstdc++', '' ) ;
%
fprintf( 1, ' ... New CLIBS argument:\n' ) ;
fprintf( 1, '     CLIBS = %s\n', Clibs ) ;

%=======================================================================
% Return additionnal argument for 'mex.m' function which removes the
% link with stdc++ library
%=======================================================================
mex_addopt = [ 'CLIBS=''', Clibs, '''' ] ;


%#######################################################################
% III] UTILITY EMBEDDED FUNCTION: "sf_setlib"
%
%#######################################################################
function  [lib, ok] = sf_setlib( path, dll_ext, static_ext, flagSTATIC )


%=======================================================================
% Outputs initialisation
%=======================================================================
[lib, ok] = deal( '', 0 ) ;

%=======================================================================
% Generate pathnames to expected libraries (dynamic or static)
%=======================================================================
% Dynamic library
dll = [ path, dll_ext ] ;
% Static library
static = [ path, static_ext ] ;

%=======================================================================
% Depending on 'flagSTATIC', start by searching dynamic or static
% library
%=======================================================================
% Start by searching dynamic library
if ( flagSTATIC < 1 )
   lib1 = dll ;
   lib2 = static ;
% Start by searching static library
else
   lib1 = static ;
   lib2 = dll ;
end

%=======================================================================
% Try to find expected libraries
%=======================================================================
if ( isequal( exist( lib1, 'file' ), 2 ) == 1 )
   lib = lib1 ;
   ok = 2 ;
elseif ( isequal( exist( lib2, 'file' ), 2 ) == 1 )
   lib = lib2 ;
   ok = 1 ;
else
   %%lib = sprintf( 'Cannot find ''%s''...\n', lib1 ) ;
   lib = []; 
end


%=======================================================================
% END OF FUNCTION
%=======================================================================
