#include <fffpy.h>
#include <fff.h>
#include <fff_vector.h>
#include <fff_gen_stats.h>


#define PY_ARRAY_UNIQUE_SYMBOL PyArray_API

#define DEFAULT_AXIS 0

/* Code pour la creation du module */ 
static char quantile_doc[] = 
" q = quantile(data, ratio=.5, interp=False, axis=0)\n\
Comments to follow";
static char mahalanobis_doc[] = 
" d2 = mahalanobis(data, vardata)\n\
data and vardata assumed C contiguous. More comments to follow";
static char module_doc[] = 
" Basic statistical routines.\n\
Author: Alexis Roche (CEA, Orsay, France), 2004-2007.";


/* 
   Quantile

*/

#define DEFAULT_RATIO 0.5
#define DEFAULT_INTERP 0 

static PyObject* quantile(PyObject* self, PyObject* args, PyObject* kw) 
{
  double ratio = DEFAULT_RATIO;
  int interp = DEFAULT_INTERP;
  int axis = DEFAULT_AXIS;
  PyArrayObject *X, *Y; 
  fff_vector *x, *y; 
  PyArrayMultiIterObject* multi;
  npy_intp ndim, dim_axis;
  npy_intp* dimAux; 
  static char *kwlist[] = {"data", "ratio", "interp", "axis", NULL}; 

  /* Parse input */ 
  if ( !PyArg_ParseTupleAndKeywords( args, kw, "O!|dii:quantile", kwlist, 
				     &PyArray_Type, &X, &ratio, &interp, &axis ) )
    Py_RETURN_NONE; 
  
  /* Get input dimensions */
  ndim = PyArray_NDIM( X ); 
  dim_axis = PyArray_DIM( X, axis );
  dimAux = (npy_intp*) malloc( ndim*sizeof(npy_intp) );
  memcpy( (void*)dimAux, (void*)PyArray_DIMS(X), ndim*sizeof(npy_intp) ); 
 
  /* Allocate output array */
  dimAux[axis] = 1;
  Y = (PyArrayObject*) PyArray_SimpleNew(ndim, dimAux, NPY_DOUBLE); 

  /* Create multi iterator */ 
  multi = (PyArrayMultiIterObject*)PyArray_MultiIterAllButAxis(2, axis, X, Y);

  /* Create fff vectors (copy or not) */ 
  x = fff_vector_fromPyArrayIter(multi->iters[0], axis); 
  y = fff_vector_fromPyArrayIter(multi->iters[1], axis); 
 
  /* DEBUG */
  fprintf(stderr, "multi->index=%d  multi->size=%d\n", multi->index, multi->size); 

  /* Loop */ 
  while(multi->index < multi->size) {

    /* DEBUG */
    fprintf(stderr, "Everything fine.\n"); 

    fff_vector_get_fromPyArrayIter(x, multi->iters[0], axis); 
    fff_vector_get_fromPyArrayIter(y, multi->iters[1], axis); 

    /* DEBUG */
    fprintf(stderr, "Everything okay.\n"); 
  
    y->data[0] = fff_vector_quantile(x, ratio, interp); 

    /* DEBUG */
    fprintf(stderr, "So fine...\n"); 

    fff_vector_const_toPyArrayIter(y, multi->iters[1], axis); 

    /* DEBUG */
    fprintf(stderr, "Hey hey... still alive!\n"); 

    PyArray_MultiIter_NEXT(multi);

  }

  /* Free memory */ 
  free(dimAux); 
  fff_vector_delete(x);
  fff_vector_delete(y);
  Py_XDECREF(multi);
  
  /* Output */
  return (PyObject*)Y;
}



static PyObject* mahalanobis(PyObject* self, PyObject* args, PyObject* kw) 
{
  PyArrayObject *X, *VX, *D2; 
  fff_vector *x, *vx, *d2; 
  PyArrayMultiIterObject* multi;
  int axis = DEFAULT_AXIS;
  npy_intp ndim, n;
  npy_intp* dimAux;  
  fff_vector *x_tmp, *vx_tmp;
  fff_matrix *Sf;  
  fff_matrix Sview; 
  static char *kwlist[] = {"data", "vardata", NULL}; 

  /* Parse input */ 
  if ( !PyArg_ParseTupleAndKeywords( args, kw, "O!O!:mahalanobis", kwlist, 
				     &PyArray_Type, &X,
				     &PyArray_Type, &VX ) )
    Py_RETURN_NONE; 
  
  
  /* Dimensions */
  n = PyArray_DIM(X, 0); 
  ndim = PyArray_NDIM(X); 
  
  /* Allocate output array */
  dimAux = (npy_intp*) malloc( ndim*sizeof(npy_intp) );
  memcpy( (void*)dimAux, (void*)PyArray_DIMS(X), ndim*sizeof(npy_intp) ); 
  dimAux[0] = 1;
  D2 = (PyArrayObject*) PyArray_SimpleNew( ndim, dimAux, NPY_DOUBLE ); 
  
  /* Allocate local structures */ 
  x_tmp = fff_vector_new(n); 
  vx_tmp = fff_vector_new(n*n); 
  Sf = fff_matrix_new(n, n); 
  
  /* Create multi iterator */ 
  multi = (PyArrayMultiIterObject*)PyArray_MultiIterAllButAxis(3, axis, X, VX, D2);
  
  /* Create fff vectors (copy or not) */ 
  x = fff_vector_fromPyArrayIter(multi->iters[0], axis); 
  vx = fff_vector_fromPyArrayIter(multi->iters[1], axis); 
  d2 = fff_vector_fromPyArrayIter(multi->iters[2], axis); 
  
  /* Loop */ 
  while(multi->index < multi->size) {

    fff_vector_get_fromPyArrayIter(x, multi->iters[0], axis); 
    fff_vector_get_fromPyArrayIter(vx, multi->iters[1], axis); 
    fff_vector_get_fromPyArrayIter(d2, multi->iters[2], axis); 

    fff_vector_memcpy(x_tmp, x); 
    fff_vector_memcpy(vx_tmp, vx); 
    Sview = fff_matrix_view(vx_tmp->data, n, n, n); /* OK because vx_tmp is contiguous */ 
    d2->data[0] = fff_mahalanobis(x_tmp, &Sview, Sf);
    
    fff_vector_const_toPyArrayIter(d2, multi->iters[2], axis); 
    PyArray_MultiIter_NEXT(multi);

  }

  /* Free memory */ 
  free( dimAux ); 
  fff_vector_delete(x_tmp); 
  fff_vector_delete(vx_tmp); 
  fff_matrix_delete(Sf); 
  Py_XDECREF(multi);
  
  /* Output */
  return (PyObject*)D2; 
}

 



static PyMethodDef module_methods[] = {
  {"quantile", (PyCFunction)quantile, METH_KEYWORDS, quantile_doc},
  {"mahalanobis", (PyCFunction)mahalanobis, METH_KEYWORDS, mahalanobis_doc},
  {NULL, NULL, 0, NULL}
};


void initroutines(void)
{
  Py_InitModule3("routines", module_methods, module_doc);
  fffpy_import_array();
  import_array();   /* required NumPy initialization */
  return; 
}
