#include "fffpy.h"


extern void _no_warning_( void )
{
  import_array(); 
  return; 
}



/* This function must be called before the module can work
   because PyArray_API is defined static, in order not to share that symbol
   within the dso. (import_array() asks the pointer value to the python process)
*/
void fffpy_import_array(void) { 
  import_array(); 
  return;
}

/* Static functions */
static fff_vector* _fff_vector_fromPtr( const char* data, npy_intp dim, npy_intp stride, int type, int isaligned ); 
static void _fff_vector_copy_fromPtr( fff_vector* y, const char* data, npy_intp stride, int type ); 
npy_intp _PyArray_find_main_axis( const PyArrayObject* x, int* ok ); 


/* 
   Get a fff_vector from an input PyArray. This function acts as a
   fff_vector constructor that is compatible with fff_vector_delete.
*/

static fff_vector* _fff_vector_fromPtr( const char* data, npy_intp dim, npy_intp stride, int type, int isaligned )
{
  fff_vector* y; 

  /* If the input array is double, just wrap without copying */
  if ( (type == NPY_DOUBLE) && isaligned ) {
    y = (fff_vector*) malloc( sizeof(fff_vector) ); 
    y->size = (size_t) dim;
    y->stride = (size_t) stride / (size_t) sizeof(double);
    y->data = (double*) data;
    y->owner = 0; 
  }
  /* Otherwise, output a owner contiguous vector with copied data */
  else {
    y = fff_vector_new( (size_t)dim ); 
    _fff_vector_copy_fromPtr( y, data, stride, type ); 
  }

  return y; 
}

/* This routine assumes that y->owner is true (and so y is contiguous) */ 
static void _fff_vector_copy_fromPtr( fff_vector* y, const char* data, npy_intp stride, int type ) 
{
  npy_intp dim[1] = {(npy_intp)y->size}; 
  npy_intp strides[1] = {stride};   
  PyArrayObject* x = (PyArrayObject*) PyArray_New( &PyArray_Type, 1, dim, type, strides, 
						   (void*)data, 0, NPY_CARRAY, NULL ); 
  PyArrayObject* xd = (PyArrayObject*) PyArray_SimpleNewFromData( 1, dim, NPY_DOUBLE, (void*)y->data );
  PyArray_CastTo( xd, x ); 
  Py_XDECREF( xd );
  Py_XDECREF( x );
  return; 
}


/* Find the axis with largest dimension */ 
npy_intp _PyArray_find_main_axis( const PyArrayObject* x, int* ok )
{
  npy_intp axis, count, i, dim, ndim = PyArray_NDIM( x ); 
  *ok = 1; 

  axis = 0; 
  count = 0; 
  for( i=0; i<ndim; i++ ) {
    dim = PyArray_DIM(x,i); 
    if ( dim > 1 ) {
      count ++; 
      axis = i;
    }
  }

  if ( count > 1 ) 
    *ok = 0; 

  return axis;
}

fff_vector* fff_vector_fromPyArray( const PyArrayObject* x ) 
{
  fff_vector* y;
  int ok; 
  npy_intp axis = _PyArray_find_main_axis( x, &ok );

  if ( !ok ) {
    fprintf(stderr, "Error 1: input array is not a vector\n");
    return NULL;
  }

  y = _fff_vector_fromPtr( PyArray_DATA(x),
                           PyArray_DIM(x, axis),
                           PyArray_STRIDE(x, axis),
                           PyArray_TYPE(x), 
			   PyArray_ISALIGNED(x) );

  return y; 
}


/*
  Export a fff_vector to a PyArray, and delete it. This function is a
  fff_vector destructor compatible with any of the following
  constructors: fff_vector_new and fff_vector_fromPyArray.
*/ 
PyArrayObject* fff_vector_toPyArray( fff_vector* y ) 
{
  if ( y == NULL ) 
    return NULL;
  PyArrayObject* x; 
  size_t size = y->size;
  npy_intp dims[1]; 

  dims[0] = (npy_intp) size;  
  
  /* If the fff_vector is owner (hence contiguous), just pass the
     buffer to Python and transfer ownership */ 
  if ( y->owner ) {
    x = (PyArrayObject*) PyArray_SimpleNewFromData( 1, dims, NPY_DOUBLE, (void*)y->data );
    x->flags = (x->flags) | NPY_OWNDATA; 
  }
  /* Otherwise, create Python array from scratch */ 
  else 
    x = fff_vector_const_toPyArray( y ); 

  /* Ciao bella */ 
  free( y );
   
  return x;
}

/* Export without deleting */ 
PyArrayObject* fff_vector_const_toPyArray( const fff_vector* y )
{
  PyArrayObject* x;
  size_t i, size = y->size, stride = y->stride; 
  double* data = (double*) malloc( size*sizeof(double) ); 
  double* bufX = data; 
  double* bufY = y->data; 
  npy_intp dims[1]; 
  
  dims[0] = (npy_intp) size;  
  for ( i=0; i<size; i++, bufX++, bufY+=stride ) 
    *bufX = *bufY; 
  x = (PyArrayObject*) PyArray_SimpleNewFromData(1, dims, NPY_DOUBLE, (void*)data);
  x->flags = (x->flags) | NPY_OWNDATA; 

  return x; 
}


/* 
   Get a fff_matrix from an input PyArray. This function acts as a
   fff_vector constructor that is compatible with fff_vector_delete.
*/ 
fff_matrix* fff_matrix_fromPyArray( const PyArrayObject* x ) 
{
  fff_matrix* y;

  /* Check that the input object is a two-dimensional array */ 
  if ( PyArray_NDIM( x ) != 2 ) {
    fprintf(stderr, "Error 3: input array is not a matrix\n");
    return NULL;
  }


  /* If the PyArray is double, contiguous and aligned just wrap without
     copying */
  if ( ( PyArray_TYPE(x) == NPY_DOUBLE ) && 
       ( PyArray_ISCONTIGUOUS(x) ) && 
       ( PyArray_ISALIGNED(x) ) ) {
    y = (fff_matrix*) malloc( sizeof(fff_matrix) ); 
    y->size1 = (size_t) PyArray_DIM(x,0);
    y->size2 = (size_t) PyArray_DIM(x,1);
    y->tda = y->size2; 
    y->data = (double*) PyArray_DATA(x);
    y->owner = 0;
  }
  /* Otherwise, output a owner (contiguous) matrix with copied
     data */
  else {
    size_t dim0 = PyArray_DIM(x,0), dim1 = PyArray_DIM(x,1);
    y = fff_matrix_new( (size_t)dim0, (size_t)dim1 ); 
    npy_intp dim[2] = {dim0, dim1}; 
    PyArrayObject* xd = (PyArrayObject*) PyArray_SimpleNewFromData( 2, dim, NPY_DOUBLE, (void*)y->data );
    PyArray_CastTo( xd, (PyArrayObject*)x ); 
    Py_XDECREF( xd );
  }
  
  return y;
}


/*
  Export a fff_matrix to a PyArray, and delete it. This function is a
  fff_matrix destructor compatible with any of the following
  constructors: fff_matrix_new and fff_matrix_fromPyArray.
*/ 
PyArrayObject* fff_matrix_toPyArray( fff_matrix* y ) 
{
  if ( y == NULL ) 
    return NULL;
  PyArrayObject* x; 
  size_t size1 = y->size1;
  size_t size2 = y->size2;
  size_t tda = y->tda; 
  npy_intp dims[2]; 

  dims[0] = (npy_intp) size1;  
  dims[1] = (npy_intp) size2;  
  
  /* If the fff_matrix is contiguous and owner, just pass the
     buffer to Python and transfer ownership */ 
  if ( ( tda == size2 ) && ( y->owner ) ) {
    x = (PyArrayObject*) PyArray_SimpleNewFromData( 2, dims, NPY_DOUBLE, (void*)y->data );
    x->flags = (x->flags) | NPY_OWNDATA; 
  }
  /* Otherwise, create PyArray from scratch. Note, the input
     fff_matrix is necessarily in row-major order. */ 
  else 
    x = fff_matrix_const_toPyArray( y ); 
  
  /* Ciao bella */ 
  free( y );
   
  return x;
}


/* Export without deleting */
PyArrayObject* fff_matrix_const_toPyArray( const fff_matrix* y )
{
  PyArrayObject* x;
  size_t size1 = y->size1, size2 = y->size2, tda = y->tda; 
  size_t i, j, pos;
  double* data = (double*) malloc( size1*size2*sizeof(double) ); 
  double* bufX = data;
  double* bufY = y->data; 
  npy_intp dims[2]; 

  dims[0] = (npy_intp) size1;  
  dims[1] = (npy_intp) size2;  
  for (i=0; i<size1; i++) {
    pos = tda*i;
    for (j=0; j<size2; j++, bufX++, pos++)
      *bufX = bufY[pos];
  }
  
  x = (PyArrayObject*) PyArray_SimpleNewFromData( 2, dims, NPY_DOUBLE, (void*)data );
  x->flags = (x->flags) | NPY_OWNDATA; 

  return x;  
}

/** Static routines **/ 






/**** fff_array interface ****/

fff_array* fff_array_fromPyArray( const PyArrayObject* x ) 
{
  char* proc = "fff_array_fromPyArray"; 
  fff_array* y;
  fff_datatype datatype; 
  size_t nbytes; 
  size_t dimX = 1, dimY = 1, dimZ = 1, dimT = 1; 
  size_t offX = 0, offY = 0, offZ = 0, offT = 0; 
  size_t ndims = (size_t)PyArray_NDIM(x);

  /* Check that the input array has less than 4 dimensions */ 
  if ( ndims > 4 ) {
    fprintf(stderr, "Error <%s>: input array has more than 4 dimensions\n", proc);
    return NULL;
  }
  /* Check that the input array is aligned */ 
  if ( ! PyArray_ISALIGNED(x) ) {
    fprintf(stderr, "Error <%s>: input array is not aligned\n", proc);
    return NULL;
  }
  /* Match the data type */
  switch ( PyArray_TYPE(x) ) {
  case NPY_UBYTE:
    datatype = FFF_UCHAR; 
    nbytes = sizeof(unsigned char); 
    break;
  case NPY_BYTE:
    datatype = FFF_SCHAR;
    nbytes = sizeof(signed char); 
    break;
  case NPY_USHORT: 
    datatype = FFF_USHORT;
    nbytes = sizeof(unsigned short); 
    break;
  case NPY_SHORT: 
    datatype = FFF_SSHORT;
    nbytes = sizeof(signed short); 
    break;
  case NPY_UINT: 
    datatype = FFF_UINT;
    nbytes = sizeof(unsigned int); 
    break;
  case NPY_INT: 
    datatype = FFF_INT;
    nbytes = sizeof(int); 
    break;
  case NPY_ULONG: 
    datatype = FFF_ULONG;
    nbytes = sizeof(unsigned long); 
    break;
  case NPY_LONG:
    datatype = FFF_LONG;
    nbytes = sizeof(long); 
    break;
  case NPY_FLOAT: 
    datatype = FFF_FLOAT;
    nbytes = sizeof(float); 
    break;
  case NPY_DOUBLE: 
    datatype = FFF_DOUBLE;
    nbytes = sizeof(double); 
    break;
  default: 
    fprintf(stderr, "Error <%s>: unrecognized data type\n", proc);
    return NULL;    
    break; 
  }

  /* Dimensions and offsets */ 
  dimX = PyArray_DIM(x, 0);
  offX = PyArray_STRIDE(x, 0)/nbytes;
  if ( ndims > 1 ) {
    dimY = PyArray_DIM(x, 1);
    offY = PyArray_STRIDE(x, 1)/nbytes;
    if ( ndims > 2 ) {
      dimZ = PyArray_DIM(x, 2);
      offZ = PyArray_STRIDE(x, 2)/nbytes;
      if ( ndims > 3 ) {
	dimT = PyArray_DIM(x, 3);
	offT = PyArray_STRIDE(x, 3)/nbytes;
      }
    }
  }

  /* Create array (owner) */ 
  y = fff_array_from_data_new( datatype, 
			       (void*) PyArray_DATA(x), 
			       dimX, dimY, dimZ, dimT, 
			       offX, offY, offZ, offT );
  
  return y;
}



PyArrayObject* fff_array_toPyArray( fff_array* y ) 
{
  if ( y == NULL ) 
    return NULL;
  char* proc = "fff_array_toPyarray"; 
  PyArrayObject* x; 
  npy_intp dims[4] = {y->dimX, y->dimY, y->dimZ, y->dimT};
  int datatype; 
  size_t nbytes; 

  /* Match data type */ 
  switch( y->datatype ) {
  case FFF_UCHAR:
    datatype = NPY_UBYTE; 
    nbytes = sizeof(unsigned char); 
    break;
  case FFF_SCHAR:
    datatype = NPY_BYTE;
    nbytes = sizeof(signed char); 
    break;
  case FFF_USHORT: 
    datatype = NPY_USHORT;
    nbytes = sizeof(unsigned short); 
    break;
  case FFF_SSHORT: 
    datatype = NPY_SHORT;
    nbytes = sizeof(signed short); 
    break;
  case FFF_UINT: 
    datatype = NPY_UINT;
    nbytes = sizeof(unsigned int); 
    break;
  case FFF_INT: 
    datatype = NPY_INT;
    nbytes = sizeof(int); 
    break;
  case FFF_ULONG: 
    datatype = NPY_ULONG;
    nbytes = sizeof(unsigned long); 
    break;
  case FFF_LONG:
    datatype = NPY_LONG;
    nbytes = sizeof(long); 
    break;
  case FFF_FLOAT: 
    datatype = NPY_FLOAT;
    nbytes = sizeof(float); 
    break;
  case FFF_DOUBLE: 
    datatype = NPY_DOUBLE;
    nbytes = sizeof(double); 
    break;
  default: 
    fprintf(stderr, "Error <%s>: unrecognized data type\n", proc);
    return NULL;    
    break; 
  }


  /* Make sure the fff array owns its data, which may require a copy */ 
  if ( ! y->owner ) {
    fff_array* yy = fff_array_new(y->datatype, y->dimX, y->dimY, y->dimZ, y->dimT); 
    fff_array_copy(yy, y); 
    fff_array_delete(y); 
    y = yy; 
  }

  /* Create a Python array from the array data (which is contiguous since it is owner) */ 
  x = (PyArrayObject*) PyArray_SimpleNewFromData(y->ndims, dims, datatype, (void*)y->data);

  /* Transfer ownership to Python */ 
  x->flags = (x->flags) | NPY_OWNDATA;
  y->owner = 0;  

  /* Delete array */ 
  fff_array_delete(y); 

  return x;
}




/* 
   Create a PyArrayMultiArrayIter so as to iterate simultaneously on
   several arrays except in one common axis.  This is a
   PyArrayMultiArrayIter constructor that I would like to see in the
   numpy C API, but for now it is "hosted" in fffPy.h

   TODO: check dimensions. 
*/ 
#include <stdarg.h>

PyObject* PyArray_MultiIterAllButAxis( int n, int axis, ... )
{
  va_list va;
  PyArrayMultiIterObject *multi;
  PyObject *current;
  PyObject *arr;
  
  int i, err=0;
  
  if (n < 2 || n > NPY_MAXARGS) {
    PyErr_Format(PyExc_ValueError,
		 "Need between 2 and (%d) "				\
		 "array objects (inclusive).", NPY_MAXARGS);
    return NULL;
  }
  
  /* fprintf(stderr, "multi new...");*/
  
  multi = (PyArrayMultiIterObject *) malloc(sizeof(PyArrayMultiIterObject));
  if (multi == NULL) return PyErr_NoMemory();
  PyObject_Init((PyObject *)multi, &PyArrayMultiIter_Type);
  
  for (i=0; i<n; i++) multi->iters[i] = NULL;
  multi->numiter = n;
  multi->index = 0;
  
  va_start(va, axis);
  for (i=0; i<n; i++) {
    current = va_arg(va, PyObject *);
    arr = PyArray_FROM_O(current);
    if (arr==NULL) {
      err=1; break;
    }
    else {
      multi->iters[i] = (PyArrayIterObject *)PyArray_IterAllButAxis(arr, &axis);
      Py_DECREF(arr);
    }
  }
  
  va_end(va);
  
  /*if (!err && PyArray_Broadcast(multi) < 0) err=1;*/ 
  
  if (err) {
    Py_DECREF(multi);
    return NULL;
  }
  
  PyArray_MultiIter_RESET(multi);
  
  return (PyObject *)multi;
}


/* Create an fff_vector from a PyArrayIter object */ 
fff_vector* fff_vector_fromPyArrayIter( const PyArrayIterObject* it, npy_intp axis )
{
  fff_vector* y; 
  char* data = PyArray_ITER_DATA( it );
  PyArrayObject* ao = (PyArrayObject*) it->ao; 
  npy_intp dim = PyArray_DIM( ao, axis ); 
  npy_intp stride = PyArray_STRIDE( ao, axis ); 
  int type = PyArray_TYPE( ao ); 
  int isaligned = PyArray_ISALIGNED( ao ); 

  y = _fff_vector_fromPtr( data, dim, stride, type, isaligned ); 

  return y;
}

/* Fetch vector data in an iterator */ 
void fff_vector_get_fromPyArrayIter( fff_vector* y, const PyArrayIterObject* it, npy_intp axis ) 
{
  /* Case owner (PyArray non-DOUBLE): copy data */ 
  if ( y->owner ) {
    PyArrayObject* ao = (PyArrayObject*) it->ao; 
    _fff_vector_copy_fromPtr( y, PyArray_ITER_DATA(it),
			      PyArray_STRIDE(ao, axis), PyArray_TYPE(ao) ); 
    
  }
  /* Case non-owner (PyArray DOUBLE): borrow data */ 
  else 
    y->data = (double*) PyArray_ITER_DATA( it ); 

  return; 
}

/* y assumed owner (otherwise, there is no point to call this function) */ 
void fff_vector_const_toPyArrayIter( const fff_vector* y, PyArrayIterObject* it, npy_intp axis ) 
{
  if ( ! y->owner )
    return; 

  npy_intp dims[1], strides[1]; 
  PyArrayObject* ao = (PyArrayObject*) it->ao; 
  dims[0] = (npy_intp) y->size;
  strides[0] = PyArray_STRIDE(ao, axis);

  /* Encapsulate y into a PyArray */  
  PyArrayObject* Py_y = (PyArrayObject*) PyArray_SimpleNewFromData( 1, dims, NPY_DOUBLE, (void*)y->data );
 
  /* Current view of the PyArray along axis */ 
  PyArrayObject* Py_view = (PyArrayObject*) PyArray_New( &PyArray_Type, 1, dims, PyArray_TYPE(ao), strides, 
							 (void*)PyArray_ITER_DATA(it), 0, NPY_BEHAVED, NULL ); 
  
  /* Copy (with cast) Py_y into Py_view */ 
  PyArray_CastTo( Py_view, Py_y ); 

  Py_XDECREF( Py_y );
  Py_XDECREF( Py_view );

  return; 
}



