#include <Python.h>
#include <numpy/arrayobject.h>
#include <fff_vector.h>
#include <fff_matrix.h>
#include <fff_array.h>


/*!
  \file fffpy.h
  \brief Python interface to \a fff 
  \author Alexis Roche, Benjamin Thyreau, Bertrand Thirion
  \date 2006
*/

#ifndef NPY_VERSION
#define npy_intp intp
#define NPY_OWNDATA OWNDATA
#define NPY_CONTIGUOUS CONTIGUOUS
#define NPY_BEHAVED BEHAVED_FLAGS
#endif

/*!
   \brief Import numpy C API

   Any Python module written in C, and using the fffpy interface, must
   call this function to work, because \c PyArray_API is defined
   static, in order not to share that symbol within the
   dso. (import_array() asks the pointer value to the python process)
*/
extern void fffpy_import_array(void);


/*!
  \brief Convert \c PyArrayObject to \c fff_vector 
  \param x input numpy array 
  
  This function may be seen as a \c fff_vector constructor compatible
  with \c fff_vector_delete. If the input has type \c PyArray_DOUBLE,
  whether or not it is contiguous, the new \c fff_vector is not
  self-owned and borrows a reference to the PyArrayObject's
  data. Otherwise, data are copied and the \c fff_vector is
  self-owned (hence contiguous) just like when created from
  scratch. Notice, the function returns \c NULL if the input array
  has more than one dimension.
*/ 
fff_vector* fff_vector_fromPyArray( const PyArrayObject* x ); 

/*!
  \brief Convert \c fff_vector to \c PyArrayObject 
  \param y input vector

  Conversely to \c fff_vector_fromPyArray, this function acts as a \c
  fff_vector destructor compatible with \c fff_vector_new, returning
  a new PyArrayObject reference. If the input vector is contiguous and
  self-owned, array ownership is simply transferred to Python;
  otherwise, the data array is copied. 
*/ 
PyArrayObject* fff_vector_toPyArray( fff_vector* y ); 

/*!
  \brief Convert \c fff_vector to \c PyArrayObject, without destruction
  \param y input const vector

  Unlike \c fff_vector_toPyArray, this function does not delete the
  input fff_vector. It always forces a copy of the data array. This
  function is useful when exporting to Python a fff_vector that
  belongs to a local structure having its own destruction method.
*/ 
PyArrayObject* fff_vector_const_toPyArray( const fff_vector* y ); 

/*!
  \brief Convert \c PyArrayObject to \c fff_matrix 
  \param x input numpy array 
  
  This function may be seen as a \c fff_matrix constructor compatible
  with \c fff_matrix_free. If the input has type \c PyArray_DOUBLE and
  is contiguous, the new \c fff_matrix is not self-owned and borrows a
  reference to the PyArrayObject's data. Otherwise, data are copied
  and the \c fff_matrix is self-owned (hence contiguous) just like
  when created from scratch. \c NULL is returned if the input array
  does not have exactly two dimensions.

  Remarks: 1) non-contiguity provokes a copy because the \c fff_matrix
  structure does not support strides; 2) matrices in column-major
  order (Fortran convention) always get copied using this function.
*/  
fff_matrix* fff_matrix_fromPyArray( const PyArrayObject* x );

/*!
  \brief Convert \c fff_matrix to \c PyArrayObject 
  \param y input matrix
  
  Conversely to \c fff_matrix_fromPyArray, this function acts as a \c
  fff_matrix destructor compatible with \c fff_matrix_new, returning
  a new PyArrayObject reference. If the input matrix is contiguous and
  self-owned, array ownership is simply transferred to Python;
  otherwise, the data array is copied. 
*/  
PyArrayObject* fff_matrix_toPyArray( fff_matrix* y ); 

/*!
  \brief Convert \c fff_matrix to \c PyArrayObject, without destruction
  \param y input const matrix

  Unlike \c fff_matrix_toPyArray, this function does not delete the
  input fff_matrix. It always forces a copy of the data array. This
  function is useful when exporting to Python a fff_matrix that
  belongs to a local structure having its own destruction method.
*/ 
PyArrayObject* fff_matrix_const_toPyArray( const fff_matrix* y ); 



/*!
  \brief Maps a numpy array to an fff_array  
  \param x input array 

  This function instantiates an fff_array that borrows data from the
  numpy array. Delete using  \c fff_array_delete.

*/
fff_array* fff_array_fromPyArray( const PyArrayObject* x ); 
PyArrayObject* fff_array_toPyArray( fff_array* y ); 


/*!
  \brief 
  \param n number of arrays
  \param axis non-iterated axis for each array 
  
  Example: 
  
  \code

  PyArrayObject *X, *Y;
  fff_vector x, y; 
  int axis;
 
  PyArrayMultiIterObject* multi = (PyArrayMultiIterObject*)PyArray_MultiIterAllButAxis(2, axis, X, Y); 
  
  while(multi->index < multi->size) {
     x = fff_vector_view(multi->iters[0]->dataptr, dim_axisX, multi->iters[0]->strides[axis]); 
     y = fff_vector_view(multi->iters[1]->dataptr, dim_axisY, multi->iters[1]->strides[axis]);
     fff_my_func(y, x); 
     PyArray_MultiIter_NEXT(multi);
  }

  Py_DECREF(multi);

  \endcode
*/ 
extern PyObject* PyArray_MultiIterAllButAxis( int n, int axis, ... ); 

extern fff_vector* fff_vector_fromPyArrayIter(const PyArrayIterObject* it, npy_intp axis); 
extern void fff_vector_get_fromPyArrayIter(fff_vector* y, const PyArrayIterObject* it, npy_intp axis);
extern void fff_vector_const_toPyArrayIter(const fff_vector* y, PyArrayIterObject* it, npy_intp axis); 
