#include <Python.h>
#include <numpy/arrayobject.h>

#include "fff_image.h"


/*!
  \file fffPy.h
  \brief Python interface to \a fff 
  \author Alexis Roche, Benjamin Thyreau, Bertrand Thirion
  \date 2006
*/

#ifndef NPY_VERSION
#define npy_intp intp
#define NPY_OWNDATA OWNDATA
#define NPY_CONTIGUOUS CONTIGUOUS
#define NPY_BEHAVED BEHAVED_FLAGS
#endif

/* This is a dummy constructor of an 0-sized vector */
#define fffPyZeroLONG()	(PyArrayObject*)PyArray_SimpleNew(1,(npy_intp*)"\0\0\0\0", PyArray_LONG);
#define fffPyZeroDOUBLE() (PyArrayObject*)PyArray_SimpleNew(1,(npy_intp*)"\0\0\0\0",PyArray_DOUBLE);


/*!  
  \struct fff_func_iterator 
  \brief Structure to iterate a vector-based function over arbitrary
  N-dimensional arrays

  This structure, which is close in spirit to numpy's universal
  functions, makes it easy to implement a C function that iterates
  over several numpy arrays simultaneously and calls a lower-level
  function whose arguments (in and out) are lists of gsl_vectors. Both
  input and output numpy arrays must have the same dimension, except
  in the direction defining the orientation of the gsl_vectors.
*/ 
typedef struct {
  npy_intp nargout; /*!< number of output numpy arrays */
  npy_intp nargin; /*!< number of input numpy arrays */
  npy_intp axis; /*!< gsl_vector direction */
  npy_intp nitems; /* number of iterations required */ 
  PyArrayIterObject** PyIter_argout; /* list of output PyArrays */ 
  PyArrayIterObject** PyIter_argin; /* list of input PyArrays */ 
  gsl_vector** gsl_argout; /* list of output gsl_vectors */ 
  gsl_vector** gsl_argin; /* list of input gsl_vectors */ 
} fff_func_iterator; 


/*!
   \brief Import numpy C API

   Any Python module written in C, and using the fffPy interface, must
   call this function to work, because \c PyArray_API is defined
   static, in order not to share that symbol within the
   dso. (import_array() asks the pointer value to the python process)
*/
int fffPy_import_array( void );

/*!
   \brief Convert \c PyArrayObject to \c gsl_vector 
   \param x input numpy array 

   This function may be seen as a \c gsl_vector constructor compatible
   with \c gsl_vector_free. If the input has type \c PyArray_DOUBLE,
   whether or not it is contiguous, the new \c gsl_vector is not
   self-owned and borrows a reference to the PyArrayObject's
   data. Otherwise, data are copied and the \c gsl_vector is
   self-owned (hence contiguous) just like when created from
   scratch. Notice, the function returns \c NULL if the input array
   has more than one dimension.
*/ 
gsl_vector* gsl_vector_fromPyArray( const PyArrayObject* x ); 

/*!
  \brief Convert \c gsl_vector to \c PyArrayObject 
  \param y input vector

  Conversely to \c gsl_vector_fromPyArray, this function acts as a \c
  gsl_vector destructor compatible with \c gsl_vector_alloc, returning
  a new PyArrayObject reference. If the input vector is contiguous and
  self-owned, array ownership is simply transferred to Python;
  otherwise, the data array is copied. 
*/ 
PyArrayObject* gsl_vector_toPyArray( gsl_vector* y ); 

/*!
  \brief Convert \c gsl_vector to \c PyArrayObject, without destruction
  \param y input const vector

  Unlike \c gsl_vector_toPyArray, this function does not delete the
  input gsl_vector. It always forces a copy of the data array. This
  function is useful when exporting to Python a gsl_vector that
  belongs to a local structure having its own destruction method.
*/ 
PyArrayObject* gsl_vector_const_toPyArray( const gsl_vector* y ); 

/*! 
  \brief Convert \c PyArrayObject to \c gsl_vector_long 
  \param x input numpy array 

  Similar to \c gsl_vector_fromPyArray. 
*/
gsl_vector_long* gsl_vector_long_fromPyArray( const PyArrayObject* x ); 
/*! 
  \brief Convert \c gsl_vector_long to \c PyArrayObject
  \param y input vector 

  Similar to \c gsl_vector_toPyArray. 
*/
PyArrayObject* gsl_vector_long_toPyArray( gsl_vector_long* y );
/*! 
  \brief Convert \c gsl_vector_long to \c PyArrayObject, without destruction. 
  \param y input vector 
  
  Similar to \c gsl_vector_const_toPyArray. 
*/
PyArrayObject* gsl_vector_long_const_toPyArray( const gsl_vector_long* y );  

/*!
  \brief Convert \c PyArrayObject to \c gsl_matrix 
  \param x input numpy array 
  
  This function may be seen as a \c gsl_matrix constructor compatible
  with \c gsl_matrix_free. If the input has type \c PyArray_DOUBLE and
  is contiguous, the new \c gsl_matrix is not self-owned and borrows a
  reference to the PyArrayObject's data. Otherwise, data are copied
  and the \c gsl_matrix is self-owned (hence contiguous) just like
  when created from scratch. \c NULL is returned if the input array
  does not have exactly two dimensions.

  Remarks: 1) non-contiguity provokes a copy because the \c gsl_matrix
  structure does not support strides; 2) matrices in column-major
  order (Fortran convention) always get copied using this function.
*/  
gsl_matrix* gsl_matrix_fromPyArray( const PyArrayObject* x );

/*!
  \brief Convert \c gsl_matrix to \c PyArrayObject 
  \param y input matrix
  
  Conversely to \c gsl_matrix_fromPyArray, this function acts as a \c
  gsl_matrix destructor compatible with \c gsl_matrix_alloc, returning
  a new PyArrayObject reference. If the input matrix is contiguous and
  self-owned, array ownership is simply transferred to Python;
  otherwise, the data array is copied. 
*/  
PyArrayObject* gsl_matrix_toPyArray( gsl_matrix* y ); 

/*!
  \brief Convert \c gsl_matrix to \c PyArrayObject, without destruction
  \param y input const matrix

  Unlike \c gsl_matrix_toPyArray, this function does not delete the
  input gsl_matrix. It always forces a copy of the data array. This
  function is useful when exporting to Python a gsl_matrix that
  belongs to a local structure having its own destruction method.
*/ 
PyArrayObject* gsl_matrix_const_toPyArray( const gsl_matrix* y ); 

/*!
  \brief Convert \c PyArrayObject to \c gsl_matrix_long 
  \param x input numpy array 

  Similar to \c gsl_matrix_fromPyArray. 
*/
gsl_matrix_long* gsl_matrix_long_fromPyArray( const PyArrayObject* x ); 

/*!
  \brief Convert \c gsl_matrix_long to \c PyArrayObject 
  \param y input matrix

  Similar to \c gsl_matrix_toPyArray. 
*/
PyArrayObject* gsl_matrix_long_toPyArray( gsl_matrix_long* y ); 

/*!
  \brief Convert \c gsl_matrix_long to \c PyArrayObject, without destruction
  \param y input matrix
  
  Similar to \c gsl_matrix_const_toPyArray. 
*/
PyArrayObject* gsl_matrix_long_const_toPyArray( const gsl_matrix_long* y ); 


/*!
   \brief Constructor for the \c fff_func_iterator structure
   \param nargout number of output numpy arrays
   \param argout list of output numpy arrays 
   \param nargin number of input numpy arrays
   \param argin list of input numpy arrays 
   \param axis vector orientation
   
   This constructor assumes that all PyArrays involved are already
   allocated, including output arrays. Moreover, they must have the
   same dimensions except in the direction given by \c axis.
*/ 
fff_func_iterator* fff_func_iterator_new( npy_intp nargout, PyArrayObject** argout, 
					  npy_intp nargin, PyArrayObject** argin,
					  npy_intp axis ); 

/*! 
   \brief Iterate a C function over several PyArrays
   \param thisone existing iterator structure 
   \param func pointer to the vector-based function to iterate
   \param par auxiliary function parameters
   
   \c func is intended to be a simple gateway function to an existing
   \a fff function. It should conform with the syntax:

   \code
   void func( gsl_vector** argout, gsl_vector** argin, void* par )
   \endcode 

   where \c argout is the list of output gsl_vectors, \c argin is the
   list of input gsl_vectors, and \c par is typically a reference to a
   pre-allocated structure in order to avoid memory allocations on
   each iteration. It is left to the developer to determine whether \c
   func assumes the correct number of input and output
   arguments. Hence, it is strongly recommended that such a gateway
   function be defined static in a Python module C file.
*/ 
void fff_func_iterator_eval( fff_func_iterator* thisone, 
			     void (*func)( gsl_vector**, gsl_vector**, void* ), 
			     void* par ); 

/*! 
   \brief Reset a \c fff_func_iterator before new loop 
   \param thisone existing iterator structure 

   This is useful when several loops over arrays are to be performed
   in the same C function.
*/ 
void fff_func_iterator_reset( fff_func_iterator* thisone ); 

/*! 
   \brief Destructor for the \c fff_func_iterator structure
   \param thisone iterator to delete

   Use after \c fff_func_iterator_new. 
*/ 
void fff_func_iterator_delete( fff_func_iterator* thisone ); 


/*!
  \brief Maps a numpy array to an fff_image  
  \param x input array 

  This function instantiates an fff_image that borrows data from the
  numpy array. Delete using  \c fff_image_delete.

*/
fff_image* fff_image_fromPyArray( const PyArrayObject* x ); 
PyArrayObject* fff_image_toPyArray( fff_image* y ); 
