#include "fff_types.h"

#include "fff_image.h"
#include "fff_macro.h"
#include "fff_iolib.h"

#include <stdlib.h>



fff_image* fff_image_from_array_new( fff_image_datatype datatype, void* buf, 
				     size_t dimX, size_t dimY, size_t dimZ, size_t dimT,
				     size_t offX, size_t offY, size_t offZ, size_t offT,
				     double sizeX, double sizeY, double sizeZ, double sizeT )
{
  fff_image* thisone; 
  fff_image_ndims ndims = fff_IMAGE_4D; 

  /* Instantiate the structure member */ 
  thisone = (fff_image*)calloc( 1, sizeof(fff_image) ); 
  if ( thisone==NULL ) 
    return NULL; 
  
  /* Decrease the number of dimensions if applicable */ 
  if ( dimT == 1 ) {
    ndims = fff_IMAGE_3D; 
    if ( dimZ == 1 ) {
      ndims = fff_IMAGE_2D;
      if ( dimY == 1 ) 
	ndims = fff_IMAGE_1D;
    }
  }
  thisone->ndims = ndims; 

  /* Set dimensions / offsets / voxel size */ 
  thisone->dimX = dimX; 
  thisone->dimY = dimY; 
  thisone->dimZ = dimZ; 
  thisone->dimT = dimT; 
  thisone->offsetX = offX; 
  thisone->offsetY = offY; 
  thisone->offsetZ = offZ; 
  thisone->offsetT = offT; 
  thisone->sizeX = sizeX; 
  thisone->sizeY = sizeY; 
  thisone->sizeZ = sizeZ; 
  thisone->sizeT = sizeT; 

  /* Set data type and point towards buffer */ 
  thisone->datatype = datatype; 
  thisone->data = buf; 
  thisone->owner = 0; 
  
  return thisone; 
}



/*

Creates a C-contiguous array. 

*/ 

fff_image* fff_image_new( fff_image_datatype datatype,
			  size_t dimX, size_t dimY, size_t dimZ, size_t dimT,
			  double sizeX, double sizeY, double sizeZ, double sizeT )
{
  char* proc = "fff_image_new"; 
  fff_image* thisone; 
  size_t nvoxels = dimX*dimY*dimZ*dimT; 
  size_t aux, offX, offY, offZ, offT;
  
  /* Offset computation */ 
  offT = 1; 
  aux = dimT; 
  offZ = aux; 
  aux *= dimZ; 
  offY = aux; 
  aux *= dimY; 
  offX = aux; 

  /* Create a new image */ 
  thisone =  fff_image_from_array_new( datatype, NULL, 
				       dimX, dimY, dimZ, dimT,
				       offX, offY, offZ, offT, 
				       sizeX, sizeY, sizeZ, sizeT ); 

  /* Gives ownership */ 
  thisone->owner = 1; 

  /* Allocate the image buffer */ 
  switch( datatype ) {

  case fff_IMAGE_UCHAR:
    {
      unsigned char* buf = (unsigned char*)calloc( nvoxels, sizeof(unsigned char) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case fff_IMAGE_SCHAR:
    {
      signed char* buf = (signed char*)calloc( nvoxels, sizeof(signed char) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case fff_IMAGE_USHORT:
    {
      unsigned short* buf = (unsigned short*)calloc( nvoxels, sizeof(unsigned short) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case fff_IMAGE_SSHORT:
    {
      signed short* buf = (signed short*)calloc( nvoxels, sizeof(signed short) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case fff_IMAGE_UINT:
    {
      unsigned int* buf = (unsigned int*)calloc( nvoxels, sizeof(unsigned int) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case fff_IMAGE_INT:
    {
      int* buf = (int*)calloc( nvoxels, sizeof(int) ); 
      thisone->data = (void*)buf;  
    }
    break;
  case fff_IMAGE_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)calloc( nvoxels, sizeof(unsigned long int) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case fff_IMAGE_LONG:
    {
      long int* buf = (long int*)calloc( nvoxels, sizeof(long int) ); 
      thisone->data = (void*)buf;  
    }
    break;
  case fff_IMAGE_FLOAT:
    {
      float* buf = (float*)calloc( nvoxels, sizeof(float) ); 
      thisone->data = (void*)buf; 
    }
    break;
  case fff_IMAGE_DOUBLE:
    {
      double* buf = (double*)calloc( nvoxels, sizeof(double) ); 
      thisone->data = (void*)buf;  
    }
    break;
  default: 
    fff_message( fff_ERROR_MSG, proc, "Unrecognized image type.\n" );
    thisone->data = NULL; 
    thisone->owner = 0; 
    break; 

  }

  return thisone; 
}


void fff_image_delete( fff_image* thisone )
{
  if ( (thisone->owner) && (thisone->data != NULL) ) 
    free(thisone->data); 
  free(thisone); 
  return; 
}



#define IMAGE_SET(val)						\
  i = 0;							\
  while ( iter.uncomplete ) {					\
    buf[i] = val;						\
    i += fff_image_iterator_update(&iter); }								

void fff_image_set( fff_image* thisone, double val )
{
  char* proc = "fff_image_set"; 
  fff_image_iterator iter = fff_image_iterator_init( thisone ); 
  size_t i;
  
  switch( thisone->datatype ) {

  case fff_IMAGE_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case fff_IMAGE_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case fff_IMAGE_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      IMAGE_SET(val);
    }
    break;
  case fff_IMAGE_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case fff_IMAGE_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case fff_IMAGE_INT:
    {
      int* buf = (int*)thisone->data;
      IMAGE_SET(val);  
    }
    break;
  case fff_IMAGE_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case fff_IMAGE_LONG:
    {
      long int* buf = (long int*)thisone->data;
      IMAGE_SET(val);  
    }
    break;
  case fff_IMAGE_FLOAT:
    {
      float* buf = (float*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  case fff_IMAGE_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      IMAGE_SET(val); 
    }
    break;
  default: 
    fff_message( fff_ERROR_MSG, proc, "Unrecognized image type.\n" );
    break; 
  }

  return; 
}


/* Return GSL_NAN if position is out of bounds */ 
double fff_image_get_value( const fff_image* thisone, 
			    size_t x, 
			    size_t y, 
			    size_t z, 
			    size_t t )
{
  double value = 0.0; 
  char* proc = "fff_image_get_value"; 
  size_t offX = thisone->offsetX; 
  size_t offY = thisone->offsetY; 
  size_t offZ = thisone->offsetZ; 
  size_t offT = thisone->offsetT; 

  /* Check coordinate range */ 
  if ( (x >= thisone->dimX) ||
       (y >= thisone->dimY) ||
       (z >= thisone->dimZ) ||
       (t >= thisone->dimT) ) 
	return GSL_NAN; 


  switch(thisone->datatype) {

  case fff_IMAGE_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf);
    }
    break;
  case fff_IMAGE_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_INT:
    {
      int* buf = (int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_LONG:
    {
      long int* buf = (long int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_FLOAT:
    {
      float* buf = (float*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      value = (double)(*buf); 
    }
    break;
  default: 
    fff_message( fff_ERROR_MSG, proc, "Unrecognized image type.\n" );
    break; 
  }
  
  return value; 
} 


void fff_image_set_value( double value, 
			  fff_image* thisone, 
			  size_t x, 
			  size_t y, 
			  size_t z, 
			  size_t t )
{
  char* proc = "fff_image_set_value"; 
  size_t offX = thisone->offsetX;
  size_t offY = thisone->offsetY; 
  size_t offZ = thisone->offsetZ; 
  size_t offT = thisone->offsetT; 

  /* Warning: coordinate range is not checked! */ 

  switch(thisone->datatype) {

  case fff_IMAGE_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (unsigned char)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (signed char)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (unsigned short)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (signed short)(fff_ROUND(value)); 
    }
    break;
  case fff_IMAGE_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (unsigned int)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_INT:
    {
      int* buf = (int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (int)(fff_ROUND(value)); 
    }
    break;
  case fff_IMAGE_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (unsigned long int)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_LONG:
    {
      long int* buf = (long int*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (long int)(fff_ROUND(value)); 
    }
    break;
  case fff_IMAGE_FLOAT:
    {
      float* buf = (float*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (float)value;
    }
    break;
  case fff_IMAGE_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      buf += x*offX + y*offY + z*offZ + t*offT; 
      *buf = (double)value;
    }
    break;
  default: 
    fff_message( fff_ERROR_MSG, proc, "Unrecognized image type.\n" );
    break; 
  }
  
  return; 
} 


double fff_image_get_value_from_index( const fff_image* thisone, 
				       size_t pos )
{
  double value = 0.0; 
  char* proc = "fff_image_get_value_from_index"; 

  /* Warning: coordinate range is not checked! */ 

  switch(thisone->datatype) {

  case fff_IMAGE_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      buf += pos;
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      buf += pos;
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      buf += pos;
      value = (double)(*buf);
    }
    break;
  case fff_IMAGE_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      buf += pos;
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      buf += pos;
      value = (double)(*buf);
    }
    break;
  case fff_IMAGE_INT:
    {
      int* buf = (int*)thisone->data;
      buf += pos;
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      buf += pos;
      value = (double)(*buf);
    }
    break;
  case fff_IMAGE_LONG:
    {
      long int* buf = (long int*)thisone->data;
      buf += pos;
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_FLOAT:
    {
      float* buf = (float*)thisone->data;
      buf += pos; 
      value = (double)(*buf); 
    }
    break;
  case fff_IMAGE_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      buf += pos; 
      value = (double)(*buf); 
    }
    break;
  default: 
    fff_message( fff_ERROR_MSG, proc, "Unrecognized image type.\n" );
    break; 
  }

  return value; 
}


void fff_image_set_value_from_index( double value, 
				     fff_image* thisone, 
				     size_t pos )
{
  char* proc = "fff_image_set_value_from_index"; 

  /* Warning: coordinate range is not checked! */ 

  switch(thisone->datatype) {

  case fff_IMAGE_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      buf += pos;
      *buf = (unsigned char)(fff_ROUND(value)); 
    }
    break;
  case fff_IMAGE_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      buf += pos;
      *buf = (signed char)(fff_ROUND(value)); 
    }
    break;
  case fff_IMAGE_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      buf += pos;
      *buf = (unsigned short)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      buf += pos;
      *buf = (signed short)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      buf += pos;
      *buf = (unsigned int)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_INT:
    {
      int* buf = (int*)thisone->data;
      buf += pos;
      *buf = (int)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      buf += pos;
      *buf = (unsigned long int)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_LONG:
    {
      long int* buf = (long int*)thisone->data;
      buf += pos;
      *buf = (long int)(fff_ROUND(value));
    }
    break;
  case fff_IMAGE_FLOAT:
    {
      float* buf = (float*)thisone->data;
      buf += pos; 
      *buf = (float)value;
    }
    break;
  case fff_IMAGE_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      buf += pos; 
      *buf = (double)value;
    }
    break;
  default: 
    fff_message( fff_ERROR_MSG, proc, "Unrecognized image type.\n" );
    break; 
  }
  
  return;
} 


fff_image fff_image_get_block( const fff_image* thisone, 
			       size_t x0, size_t x1, 
			       size_t y0, size_t y1, 
			       size_t z0, size_t z1, 
			       size_t t0, size_t t1, 
			       size_t fX, size_t fY, size_t fZ, size_t fT )
{
  char* proc = "fff_image_get_block"; 
  fff_image res; 
  fff_image_ndims ndims = fff_IMAGE_4D; 
  size_t offX = thisone->offsetX; 
  size_t offY = thisone->offsetY; 
  size_t offZ = thisone->offsetZ; 
  size_t offT = thisone->offsetT; 

  /* Warning: coordinate range is not checked! */ 
  
  if ( t1 == t0 ) {
    ndims = fff_IMAGE_3D; 
    if ( z1 == z0 ) {
      ndims = fff_IMAGE_2D;
      if ( y1 == y0 ) 
	ndims = fff_IMAGE_1D;
    }
  }
  res.ndims = ndims; 

  res.dimX = (x1-x0)/fX + 1; 
  res.dimY = (y1-y0)/fY + 1; 
  res.dimZ = (z1-z0)/fZ + 1;
  res.dimT = (t1-t0)/fT + 1; 

  res.offsetX = offX * fX; 
  res.offsetY = offY * fY; 
  res.offsetZ = offZ * fZ; 
  res.offsetT = offT * fT; 

  res.sizeX = thisone->sizeX * fX;
  res.sizeY = thisone->sizeY * fY;
  res.sizeZ = thisone->sizeZ * fZ;
  res.sizeT = thisone->sizeT * fT;

  res.datatype = thisone->datatype; 

  /* Data pointer */ 
  res.owner = 0; 
  res.data = NULL; 
  switch(thisone->datatype) {
    
  case fff_IMAGE_UCHAR:
    {
      unsigned char* buf = (unsigned char*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case fff_IMAGE_SCHAR:
    {
      signed char* buf = (signed char*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case fff_IMAGE_USHORT:
    {
      unsigned short* buf = (unsigned short*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case fff_IMAGE_SSHORT:
    {
      signed short* buf = (signed short*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf; 
    }
    break;
  case fff_IMAGE_UINT:
    {
      unsigned int* buf = (unsigned int*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case fff_IMAGE_INT:
    {
      int* buf = (int*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf; 
    }
    break;
  case fff_IMAGE_ULONG:
    {
      unsigned long int* buf = (unsigned long int*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case fff_IMAGE_LONG:
    {
      long int* buf = (long int*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf; 
    }
    break;
  case fff_IMAGE_FLOAT:
    {
      float* buf = (float*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf;
    }
    break;
  case fff_IMAGE_DOUBLE:
    {
      double* buf = (double*)thisone->data;
      buf += x0*offX + y0*offY + z0*offZ + t0*offT; 
      res.data = (void*)buf; 
    }
    break;
  default: 
    fff_message( fff_ERROR_MSG, proc, "Unrecognized image type.\n" );
    break; 
  }
  
  return res; 
}



int fff_image_has_integer_data ( const fff_image* thisone )
{
  int ok; 

  switch ( thisone->datatype ) {

  default:
    ok = 0; 
    break; 

  case fff_IMAGE_UCHAR:
  case fff_IMAGE_SCHAR:
  case fff_IMAGE_USHORT:
  case fff_IMAGE_SSHORT:
  case fff_IMAGE_UINT:
  case fff_IMAGE_INT:
  case fff_IMAGE_ULONG:
  case fff_IMAGE_LONG:
    ok = 1; 
    break;
  
  }

  return ok; 
}

void fff_image_extrema ( double* min, double* max, const fff_image* thisone ) 
{
  double val; 
  fff_image_iterator iter = fff_image_iterator_init( thisone ); 
  size_t i;
  
  /* Initialization */ 
  *min = 0.0; 
  *max = 0.0; 

  i = 0; 
  while ( iter.uncomplete ) {
    val = fff_image_get_value_from_index( thisone, i ); 
    if ( val < *min )
      *min = val; 
    else if ( val > *max ) 
      *max = val;
    i += fff_image_iterator_update(&iter); 
  }

  return; 
}


/*
  Applies an affine correction to the input image so that:

  s0 --> r0 
  s1 --> r1

*/

void fff_image_copy( fff_image* imRes, const fff_image* imSrc, 
		     double r0, double s0, 
		     double r1, double s1 )
{
  char* proc = "fff_image_copy"; 

  /* Check dimensions */ 
  if ( (imRes->dimX != imSrc->dimX) ||
       (imRes->dimY != imSrc->dimY) ||
       (imRes->dimZ != imSrc->dimZ) ||
       (imRes->dimT != imSrc->dimT) ) {
    fff_message( fff_ERROR_MSG, proc, "Images have different sizes.\n" );  
    return; 
  }
  
  fff_image_iterator itSrc = fff_image_iterator_init( imSrc ); 
  fff_image_iterator itRes = fff_image_iterator_init( imRes ); 
  size_t i, j;
  double a, b, valSrc; 

  a = (r1-r0) / (s1-s0); 
  b = r0 - a*s0; 

  i = 0; 
  j = 0; 
  while ( itSrc.uncomplete ) {
    
    valSrc = fff_image_get_value_from_index( imSrc, i ); 
    fff_image_set_value_from_index( a*valSrc+b, imRes, j ); 
    i += fff_image_iterator_update( &itSrc ); 
    j += fff_image_iterator_update( &itRes ); 

  }

  return;
}






fff_image_iterator fff_image_iterator_init(const fff_image* im)
{
  fff_image_iterator thisone;
  size_t pX, pY, pZ, pT;  
  thisone.x = 0; 
  thisone.y = 0; 
  thisone.z = 0; 
  thisone.t = 0; 
  thisone.ddimX = im->dimX - 1;
  thisone.ddimY = im->dimY - 1; 
  thisone.ddimZ = im->dimZ - 1; 
  thisone.ddimT = im->dimT - 1;
  pX = thisone.ddimX * im->offsetX; 
  pY = thisone.ddimY * im->offsetY; 
  pZ = thisone.ddimZ * im->offsetZ; 
  pT = thisone.ddimT * im->offsetT; 
  thisone.incT = im->offsetT; 
  thisone.incZ = im->offsetZ - pT; 
  thisone.incY = im->offsetY - pZ - pT; 
  thisone.incX = im->offsetX - pY - pZ - pT; 
  thisone.axis = -1; 
  thisone.uncomplete = 1; 
  return thisone; 
}

fff_image_iterator fff_image_iterator_init_skip_axis(const fff_image* im, int axis)
{
  size_t pX, pY, pZ, pT;  
  fff_image_iterator thisone = fff_image_iterator_init(im);  
  if ( (axis>=0) && (axis<4) )
    thisone.axis = axis; 
  pX = thisone.ddimX * im->offsetX; 
  pY = thisone.ddimY * im->offsetY; 
  pZ = thisone.ddimZ * im->offsetZ; 
  pT = thisone.ddimT * im->offsetT; 
  if (axis == 3)
    pT = 0; 
  else if (axis == 2)
    pZ = 0; 
  else if (axis == 1) 
    pY = 0; 
  else /* axis == 0 */ 
    pX = 0; 
  thisone.incT = im->offsetT; 
  thisone.incZ = im->offsetZ - pT; 
  thisone.incY = im->offsetY - pZ - pT; 
  thisone.incX = im->offsetX - pY - pZ - pT; 
  return thisone;
}



/* Return the index offset to access next element */ 
size_t fff_image_iterator_update( fff_image_iterator* thisone )
{
  int axis = thisone->axis; 

  if ( (thisone->t < thisone->ddimT) && (axis!=3) ) {
    thisone->t ++; 
    return thisone->incT; 
  }

  if ( (thisone->z < thisone->ddimZ) && (axis!=2) ) {
    thisone->t = 0; 
    thisone->z ++;
    return thisone->incZ; 
  }  

  if ( (thisone->y < thisone->ddimY) && (axis!=1) ) {
    thisone->t = 0; 
    thisone->z = 0; 
    thisone->y ++; 
    return thisone->incY;
  } 

 if ( (thisone->x < thisone->ddimX) && (axis!=0) ) {
    thisone->t = 0; 
    thisone->z = 0; 
    thisone->y = 0; 
    thisone->x ++; 
    return thisone->incX; 
  } 

  /* We are here because one can't move anymore */ 
  thisone->t = 0; 
  thisone->z = 0; 
  thisone->y = 0; 
  thisone->x = 0; 
  thisone->uncomplete = 0; 
  return 0; 
}



/* Image must be in DOUBLE format */ 
void fff_image_iterate_vector_function( fff_image* im, int axis, void(*func)(gsl_vector*, void*), void* par )
{
  char* proc = "fff_image_iterate_vector_function"; 
  fff_image_iterator iter; 
  size_t i, size, stride; 
  double* im_data = (double*)im->data; 
  gsl_vector x; 

  if ( im->datatype != fff_IMAGE_DOUBLE ) {
    fff_message( fff_ERROR_MSG, proc, "Image type must be double.\n" );
    return; 
  }
  if ( (axis>3) || (axis<0) ) {
    fff_message( fff_ERROR_MSG, proc, "Invalid axis.\n" );
    return; 
 }
   
  size = fff_IMAGE_GET_DIM( im, axis );
  stride = fff_IMAGE_GET_OFFSET( im, axis );

  x.size = size; 
  x.stride = stride; 
  x.owner = 0; 
  x.block = NULL; 

  iter = fff_image_iterator_init_skip_axis( im, axis );
  i = 0; 
  while ( iter.uncomplete ) {
    x.data = im_data + i;
    (*func)( &x, par );  
    i += fff_image_iterator_update(&iter); 
  }
  
  return; 
}



fff_image_datatype fff_get_image_datatype( unsigned int sizeType, unsigned int integerType, unsigned int signedType )
{
  fff_image_datatype type = fff_IMAGE_UNKNOWN_TYPE; 

  /* Case: integer type */ 
  if ( integerType ) {

    if ( signedType ) {
      if ( sizeType == sizeof(signed char) ) 
	type = fff_IMAGE_SCHAR;
      else if ( sizeType == sizeof(signed short) )
	type = fff_IMAGE_SSHORT;
      else if ( sizeType == sizeof(int) )
	type = fff_IMAGE_INT;
      else if ( sizeType == sizeof(signed long int) )
	type = fff_IMAGE_LONG;
    }
    else {
      if ( sizeType == sizeof(unsigned char) ) 
	type = fff_IMAGE_UCHAR;
      else if ( sizeType == sizeof(unsigned short) )
	type = fff_IMAGE_USHORT;
      else if ( sizeType == sizeof(unsigned int) )
	type = fff_IMAGE_UINT;
      else if ( sizeType == sizeof(unsigned long int) )
	type = fff_IMAGE_ULONG;
    }
  
  }
  
  /* Case: floating type */ 
  else {
    if ( sizeType == sizeof(float) ) 
      type = fff_IMAGE_FLOAT;
    else if ( sizeType == sizeof(double) )
      type = fff_IMAGE_DOUBLE;
  }

  return type; 

}



/* 
  Convert image values to [0,clamp-1]; typically clamp = 256.
  Possibly modify the dynamic range if the input value is
  overestimated.  For instance, the reconstructed MRI signal is
  generally encoded in 12 bits (values ranging from 0 to
  4095). Therefore, this operation may result in a loss of
  information.
*/

void fff_image_clamp( fff_image* imRes, const fff_image* imSrc, double th, int* clamp )
{
  char* proc = "fff_image_clamp"; 
  double imin, imax, tth; 
  int dmax = *clamp - 1;
 
  /* Compute input image min and max */
  fff_image_extrema( &imin, &imax, imSrc ); 

  /* Make sure the threshold is not below the min intensity */ 
  tth = GSL_MAX(th, imin);

  /* Test */ 
  if (tth>imax) {
    fff_message( fff_WARNING_MSG, proc, "Inconsistent threshold, ignored.\n" );
    tth = imin; 
  }    

  /* If the image dynamic is small, no need for compression: just
     downshift image values and re-estimate the dynamic range (hence
     imax is translated to imax-tth casted to SSHORT) */
  if ( ( fff_image_has_integer_data(imSrc) ) && ( (imax-tth)<=dmax ) ) {
    fff_image_copy( imRes, imSrc, 0, tth, 1, tth+1 ); 
    *clamp = (int)(imax-tth) + 1;
  }
 
  /* Otherwise, compress after downshifting image values (values equal
     to the threshold are reset to zero) */ 
  else 
    fff_image_copy( imRes, imSrc, 0, tth, dmax, imax ); 

  return;
}
