/*!
  \file fff_image.h
  \brief Basic image object
  \author Alexis Roche
  \date 2005-2006

  This library implements a generic 4-dimensional array object that
  can be used to represent images. 
*/

 
#ifndef fff_IMAGE
#define fff_IMAGE

#ifdef __cplusplus
extern "C" {
#endif

  /* #include <stddef.h> */ 
#include "fff_types.h"

#define fff_IMAGE_GET_DIM( im, axis )					\
  ( (axis)==0 ? (im->dimX) : ( (axis)==1 ? (im->dimY) : ( (axis)==2 ? (im->dimZ) : (im->dimT) )  )  )  
#define fff_IMAGE_GET_OFFSET( im, axis )				\
  ( (axis)==0 ? (im->offsetX) : ( (axis)==1 ? (im->offsetY) : ( (axis)==2 ? (im->offsetZ) : (im->offsetT) )  )  )  


  /*!
    \typedef fff_image_ndims
    \brief Image flag type
  */
  typedef enum {
    fff_IMAGE_1D = 1,   /*!< 1d image */
    fff_IMAGE_2D = 2,   /*!< 2d image */
    fff_IMAGE_3D = 3,   /*!< 3d image */
    fff_IMAGE_4D = 4    /*!< 4d image */
  } fff_image_ndims;
  
  /*!
    \typedef fff_image_datatype
    \brief Image data encoding type
  */
  typedef enum {
    fff_IMAGE_UNKNOWN_TYPE = -1,  /*!< unknown type */
    fff_IMAGE_UCHAR = 0,          /*!< unsigned char */
    fff_IMAGE_SCHAR = 1,          /*!< signed char */
    fff_IMAGE_USHORT = 2,         /*!< unsigned short */
    fff_IMAGE_SSHORT = 3,         /*!< signed short */
    fff_IMAGE_UINT = 4,           /*!< unsigned int */
    fff_IMAGE_INT = 5,            /*!< (signed) int */
    fff_IMAGE_ULONG = 6,          /*!< unsigned long int */
    fff_IMAGE_LONG = 7,           /*!< (signed) long int */
    fff_IMAGE_FLOAT = 8,          /*!< float */
    fff_IMAGE_DOUBLE = 9          /*!< double */
  } fff_image_datatype;


  /*!
    \brief Return the data type that matches given features
    \param sizeType size in bytes
    \param integerType if zero, a floating-point type (\c float or \c double) is assumed
    \param signedType for integer types, tells whether the type is signed or not
  */
  extern fff_image_datatype fff_get_image_datatype( unsigned int sizeType, 
						    unsigned int integerType, 
						    unsigned int signedType ); 
  
  /*!
    \struct fff_image
    \brief The fff image structure
    
    Image values are stored in a \c void linear array, the actual
    encoding type being specified by the field \c datatype. The image
    dimension along each axis are encoded by fields starting with \c
    dim, while the \c ndims flag specifies the biggest axis index
    corresponding to a non-unitary dimension; it essentially defines
    whether the image is 1d, 2d, 3d, or 4d. The use of offsets (or
    strides) makes the object independent from any storage
    convention. A pixel with coordinates (\a x, \a y, \a z, \a t) may
    be accessed using a command like:

    \code 
    value = im->data[ x*im->offsetX + y*im->offsetY + z*im->offsetZ + t*im->offsetT ];
    \endcode

    Note that this approach makes it possible to extract a sub-image
    from an original image without the need to reallocate memory. 
  */
  typedef struct {
    fff_image_ndims ndims; /*!< Image flag */
    fff_image_datatype datatype; /*!< Image encoding type */
    size_t dimX; /*!< Dimension (number of pixels) along first axis */
    size_t dimY; /*!< Dimension (number of pixels) along second axis */
    size_t dimZ; /*!< Dimension (number of pixels) along third axis */
    size_t dimT; /*!< Dimension (number of pixels) along fourth axis */
    size_t offsetX; /*!< Offset along first axis */
    size_t offsetY; /*!< Offset along second axis */
    size_t offsetZ; /*!< Offset along third axis */
    size_t offsetT; /*!< Offset along fourth axis */
    double sizeX; /*!< Pixel size along first axis */
    double sizeY; /*!< Pixel size along second axis */
    double sizeZ; /*!< Pixel size along third axis */
    double sizeT; /*!< Pixel size along fourth axis */
    void* data; /*!< Image buffer */
    int owner; /*!< Non-zero if the object owns its data */
  } fff_image;


  /*!
    \struct fff_image_iterator
    \brief Image iterator structure
  */
  typedef struct {
    size_t x; 
    size_t y; 
    size_t z; 
    size_t t;
    size_t ddimX; 
    size_t ddimY; 
    size_t ddimZ; 
    size_t ddimT;
    size_t incX; 
    size_t incY; 
    size_t incZ; 
    size_t incT; 
    int axis; 
    int uncomplete; 
  } fff_image_iterator;
    

  /*! 
    \brief Constructor for the fff_image structure 
    \param datatype image encoding type
    \param dimX number of pixels along the first axis
    \param dimY number of pixels along the second axis
    \param dimZ number of pixels along the third axis
    \param dimT number of pixels along the fourth axis
    \param sizeX pixel size along the first axis
    \param sizeY pixel size along the second axis
    \param sizeZ pixel size along the third axis
    \param sizeT pixel size along the fourth axis

    This function allocates a new image buffer. 
  */
  extern fff_image* fff_image_new( fff_image_datatype datatype,
				   size_t dimX, 
				   size_t dimY, 
				   size_t dimZ, 
				   size_t dimT,
				   double sizeX, 
				   double sizeY, 
				   double sizeZ, 
				   double sizeT ); 
  
  /*!
    \brief Destructor for the \c fff_image structure
    \param thisone fff_image member to be deleted
  */
  extern void fff_image_delete( fff_image* thisone );
  
  
  
  /*! 
    \brief Another constructor for the fff_image structure 
    \param datatype image encoding type
    \param buf already allocated image buffer 
    \param dimX number of pixels along the first axis
    \param dimY number of pixels along the second axis
    \param dimZ number of pixels along the third axis
    \param dimT number of pixels along the fourth axis
    \param offX offset along the first axis
    \param offY offset along the second axis
    \param offZ offset along the third axis
    \param offT offset along the fourth axis
    \param sizeX pixel size along the first axis
    \param sizeY pixel size along the second axis
    \param sizeZ pixel size along the third axis
    \param sizeT pixel size along the fourth axis

    This function assumes the image buffer is already allocated. 
  */
  extern fff_image* fff_image_from_array_new( fff_image_datatype datatype, void* buf, 
					      size_t dimX, size_t dimY, size_t dimZ, size_t dimT,
					      size_t offX, size_t offY, size_t offZ, size_t offT,
					      double sizeX, double sizeY, double sizeZ, double sizeT ); 

  /*!
    \brief Set all pixel values to a given constant
    \param thisone image
    \param c constant
  */
  extern void fff_image_set( fff_image* thisone, double c ); 

  /*! 
    \brief Generic function to access a voxel's value
    \param thisone input image
    \param x first coordinate
    \param y second coordinate
    \param z third coordinate
    \param t fourth coordinate
	
	Get image value at a specific location defined by voxel coordinates. 
	Return \c GSL_NAN if the position is out of bounds. 
 */ 
  extern double fff_image_get_value( const fff_image* thisone, 
				     size_t x, 
				     size_t y, 
				     size_t z, 
				     size_t t ); 
  /*! 
    \brief Access voxel's value through its index in the image buffer
    \param thisone input image
    \param pos index of the considered point in the image buffer address

    This function is faster than \c fff_image_get_value, but don't use
    it unless you are perfectly sure about what you are doing! 
  */ 
  extern double fff_image_get_value_from_index( const fff_image* thisone, 
						size_t pos );
  /*! 
    \brief Generic function to set one voxel's value
    \param value value to set 
    \param thisone input image
    \param x first coordinate
    \param y second coordinate
    \param z third coordinate
    \param t fourth coordinate
  */ 
  extern void fff_image_set_value( double value, 
				   fff_image* thisone, 
				   size_t x, 
				   size_t y, 
				   size_t z, 
				   size_t t ); 
  /*! 
    \brief Set one voxel's value through its index in the image buffer
    \param value value to set 
    \param thisone input image
    \param pos index of the considered point in the image buffer address

    This function is faster than \c fff_image_get_value, but don't use
    it unless you are perfectly sure about what you are doing! 
  */ 
  extern void fff_image_set_value_from_index( double value, 
					      fff_image* thisone, 
					      size_t pos );

  /*! 
    \brief Extract an image block 
    \param thisone input image
    \param x0 first coordinate of the starting point
    \param x1 first coordinate of the finishing point
    \param y0 second coordinate of the starting point
    \param y1 second coordinate of the finishing point
    \param z0 third coordinate of the starting point
    \param z1 third coordinate of the finishing point
    \param t0 fourth coordinate of the starting point
    \param t1 fourth coordinate of the finishing point
    \param fX subsampling factor in the first direction
    \param fY subsampling factor in the second direction
    \param fZ subsampling factor in the third direction
    \param fT subsampling factor in the fourth direction
  */ 
  extern fff_image fff_image_get_block( const fff_image* thisone, 
					size_t x0, size_t x1, 
					size_t y0, size_t y1, 
					size_t z0, size_t z1, 
					size_t t0, size_t t1,  
					size_t fX, size_t fY, size_t fZ, size_t fT ); 
  

  extern int fff_image_has_integer_data ( const fff_image* thisone );
  
  extern void fff_image_extrema ( double* min, double* max, const fff_image* thisone );

  extern void fff_image_copy( fff_image* imRes, const fff_image* imSrc,
			      double r0, double s0, 
			      double r1, double s1 ); 

/* 
  Convert image values to [0,clamp-1]; typically clamp = 256.
  Possibly modify the dynamic range if the input value is
  overestimated.  For instance, the reconstructed MRI signal is
  generally encoded in 12 bits (values ranging from 0 to
  4095). Therefore, this operation may result in a loss of
  information.
*/
  extern void fff_image_clamp( fff_image* imRes, const fff_image* imSrc, double th, int* clamp ); 

  extern fff_image_iterator fff_image_iterator_init(const fff_image* im); 
  extern fff_image_iterator fff_image_iterator_init_skip_axis(const fff_image* im, int axis); 
  
  extern size_t fff_image_iterator_update( fff_image_iterator* thisone ); 
  extern void fff_image_iterate_vector_function( fff_image* im, int axis, 
						 void(*func)(gsl_vector*, void*), void* par ); 
 

#ifdef __cplusplus
}
#endif

#endif
