#include "fff_routines.h"

#include "fff_types.h"

#include "fff_macro.h"
#include "fff_iolib.h"

#include <stdlib.h>


/* Declaration of static functions */ 
static double _fff_pth_element( double* x, size_t p, size_t stride, size_t size ); 
static void _fff_pth_interval( double* am, double* aM, double* x, size_t p, size_t stride, size_t size ); 

/*** STATIC FUNCTIONS ***/ 
/* BEWARE: the input array x gets modified! */ 

/*
  Pick up the sample value a so that:
  (p+1) sample values are <= a AND the remaining sample values are >= a 

*/ 

#define SWAP(a, b)  {tmp=(a); (a)=(b); (b)=tmp;}
static double _fff_pth_element( double* x, size_t p, size_t stride, size_t n )
{
  double a, tmp;  
  double *bufl, *bufr;
  size_t i, j, il, jr, stop1, stop2;
  int same_extremities;
   
  stop1 = 0; 
  il = 0; 
  jr = n-1;
  while ( stop1 == 0 ) {
    
    same_extremities = 0; 
    bufl = x + stride*il; 
    bufr = x + stride*jr; 
    if ( *bufl > *bufr )
      SWAP( *bufl, *bufr )
    else if ( *bufl == *bufr ) 
      same_extremities = 1;
    a = *bufl; 
    
    if ( il == jr ) 
      return a;
    bufl += stride; 
    i = il + 1; 
    j = jr; 
    
    stop2 = 0; 
    while ( stop2 == 0 ) {
      while ( *bufl < a ) {
	i ++;
	bufl += stride;
      }
      while ( *bufr > a ) {
	j --; 
	bufr -= stride; 
      }
      if ( j <= i ) 
	stop2 = 1; 
      else {
	SWAP( *bufl, *bufr )
	j --; bufr -= stride;
	i ++; bufl += stride; 
      } 
      
      /* Avoids infinite loops in samples with redundant values. 
         This situation can only occur with i == j */ 
      if ( (same_extremities) && (j==jr) ) {
	j --; 
	bufr -= stride; 
	SWAP(x[il*stride], *bufr) 
	stop2 = 1; 
      }
    }   
    
    /* At this point, we know that il <= j <= i; moreover: 
         if k <= j, x(j) <= a and if k > j, x(j) >= a 
         if k < i, x(i) <= a and if k >= i, x(i) >= a 
      
       We hence have: (j+1) values <= a and the remaining (n-j-1) >= a 
                        i values <= a and the remaining (n-i) >= a                        
    */
  
    if ( j > p ) 
      jr = j;
    else if ( j < p )
      il = i;
    else /* j == p */
      stop1 = 1;

  }

  return a; 
}


/* BEWARE: the input array x gets modified! */ 
static void _fff_pth_interval( double* am, double* aM, 
			       double* x, size_t p, size_t stride, size_t n )
{
  double a, tmp;
  double *bufl, *bufr; 
  size_t i, j, il, jr, stop1, stop2, stop3;
  size_t pp = p+1; 
  int same_extremities = 0; 

  *am = 0.0; 
  *aM = 0.0; 
  stop1 = 0; 
  stop2 = 0; 
  il = 0; 
  jr = n-1; 
  while ( (stop1 == 0) || (stop2 == 0) ) {

    same_extremities = 0; 
    bufl = x + stride*il; 
    bufr = x + stride*jr; 
    if ( *bufl > *bufr )
      SWAP( *bufl, *bufr ) 
    else if ( *bufl == *bufr ) 
      same_extremities = 1;
    a = *bufl;
    
    if ( il == jr ) { 
      *am=a; 
      *aM=a; 
      return;
    } 
    
    bufl += stride; 
    i = il + 1; 
    j = jr; 

    stop3 = 0; 
    while ( stop3 == 0 ) {

      while ( *bufl < a ) {
	i ++;
	bufl += stride;
      }
      while ( *bufr > a ) {
	j --; 
	bufr -= stride; 
      }
      if ( j <= i )  
	stop3 = 1; 
      else {
	SWAP( *bufl, *bufr )
	j --; bufr -= stride;
	i ++; bufl += stride; 
      } 

      /* Avoids infinite loops in samples with redundant values */ 
      if ( (same_extremities) && (j==jr) ) {
	j --; 
	bufr -= stride; 
	SWAP(x[il*stride], *bufr) 
        stop3 = 1; 
      }

    }

    /* At this point, we know that there are (j+1) datapoints <=a
       including a itself, and another (n-j-1) datapoints >=a */
    if ( j > pp ) 
      jr = j;
    else if ( j < p ) 
      il = i;
    /* Case: found percentile at p */
    else if ( j == p ) { 
      il = i; 
      *am = a; 
      stop1 = 1; 
    }
    /* Case: found percentile at (p+1), ie j==(p+1) */ 
    else { 
      jr = j; 
      *aM = a;
      stop2 = 1;
    }

  }

  return;  
}


/* 
   Given a sample x, this function computes a value q so that the
   number of sample values that are greater or equal to q is smaller
   or equal to (1-r) * sample size. 

*/ 

double fff_quantile_from_temp_data( gsl_vector* x, double r, int interp )
{
    char * proc = "fff_quantile_from_temp_data";
  if ( (r<0) || (r>1) ) {
    fff_message( fff_ERROR_MSG, proc, "Fraction number must be between 0 and 1.\n" ); 
    return 0.0; 
    }
  double m, pp; 
  double* data;
  size_t p, stride, size;
  data = (double*) PyArray_DATA(x);
  size = x->dimensions[0];
  stride = *(PyArray_STRIDES(x)) / sizeof(double);

  if ( size == 1 ) 
      return data[0];   
  
  /* Find the smallest index p so that p >= r * size */ 
  if ( !interp ) {
    pp = r * size; 
    p = fff_UNSIGNED_CEIL(pp); 
    if ( p == size ) 
      return GSL_POSINF; 
    m = _fff_pth_element( data, p, stride, size );
  }
  else {
    double wm, wM; 
    pp = r * (size-1); 
    p = fff_UNSIGNED_FLOOR(pp);
    wM = pp - (double)p;
    wm = 1.0 - wM; 
    if ( wM <= 0 ) 
      m = _fff_pth_element( data, p, stride, size );
    else { 
      double am, aM; 
      _fff_pth_interval( &am, &aM, data, p, stride, size );
      m = wm*am + wM*aM; 
    }
  }
  
  return m; 
}

static void _quantile( gsl_vector** argout, gsl_vector** argin, void* par )
{
  quantile_struct* Params = (quantile_struct*) par;
  //gsl_vector_memcpy( Params->tmp, argin[0] ); 
  argout[0]->data[0] =  fff_quantile_from_temp_data( argin[0],//Params->tmp, 
						     Params->ratio, 
						     Params->interp ); 
  return; 
} 


void fff_func_iterator_eval( fff_func_iterator* thisone, 
			     void (*func)( gsl_vector**, gsl_vector**, void* ), 
			     void* par )
{
#if 0
  npy_intp i; 
  npy_intp nargout = thisone->nargout, nargin = thisone->nargin, axis = thisone->axis; 
  PyArrayIterObject** PyIter_argout = thisone->PyIter_argout; 
  gsl_vector** gsl_argout = thisone->gsl_argout; 
  PyArrayIterObject** PyIter_argin = thisone->PyIter_argin; 
  gsl_vector** gsl_argin = thisone->gsl_argin; 
  PyArrayIterObject *PyIter_ptr; 
  gsl_vector* gsl_ptr;
 
  /* Loop over items */ 
  while( PyIter_argin[0]->index < thisone->nitems ) {

    /* Do the job for current item */
    func( gsl_argout, gsl_argin, par ); 

    /* Prepare next item */ 
    for( i=0; i<nargout; i++ ) {
      PyIter_ptr = PyIter_argout[i];
      gsl_ptr = gsl_argout[i]; 
      
      /* If the gsl vector is not a view (self-owner), then push back its content */
      if ( gsl_ptr->owner ) 	
	_gsl_vector_toPyArrayIter_update( gsl_ptr, PyIter_ptr, axis ); 
      
      PyArray_ITER_NEXT( PyIter_ptr );
      _gsl_vector_fromPyArrayIter_update( gsl_ptr, PyIter_ptr, axis ); 
    }
    for( i=0; i<nargin; i++ ) {
      PyIter_ptr = PyIter_argin[i];
      PyArray_ITER_NEXT( PyIter_ptr );
      _gsl_vector_fromPyArrayIter_update( gsl_argin[i], PyIter_ptr, axis ); 
    }
    
  } /* End loop over items */ 
  
  return;
#endif
}



fff_func_iterator* fff_func_iterator_new( npy_intp nargout, PyArrayObject** argout, 
					  npy_intp nargin, PyArrayObject** argin,
					  npy_intp npy_axis )
{
#if 0
  fff_func_iterator* thisone;
  npy_intp i; 
  PyArrayObject *Py_ptr; 
  int axis = (int)npy_axis; /* For compatibility with the new PyArray_IterAllButAxis API */ 

  /* Create this one */ 
  thisone = (fff_func_iterator*)malloc( sizeof(fff_func_iterator) );   
  if ( thisone == NULL ) 
    return NULL; 
  
  /* Number of args and axis */ 
  thisone->nargin = nargin; 
  thisone->nargout = nargout; 
  thisone->axis = axis; 

  /* Number of loop items */ 
  /* TODO: add consistency check -- all arrays should have same dim except in the axis direction */
  Py_ptr = argin[0]; 
  thisone->nitems = PyArray_SIZE(Py_ptr) / PyArray_DIM(Py_ptr, axis);
  
  /* Allocate local objects (array iterators and local gsl vectors)*/
  /** Output **/ 
  thisone->PyIter_argout = (PyArrayIterObject**)malloc( nargout*sizeof(PyArrayIterObject*) ); 
  thisone->gsl_argout = (gsl_vector**)malloc( nargout*sizeof(gsl_vector*) ); 
  for( i=0; i<nargout; i++ ) {
    thisone->PyIter_argout[i] = (PyArrayIterObject*)PyArray_IterAllButAxis( (PyObject*)argout[i], &axis );
    thisone->gsl_argout[i] = _gsl_vector_fromPyArrayIter( thisone->PyIter_argout[i], axis );
  }
  /** Input arrays **/ 
  thisone->PyIter_argin = (PyArrayIterObject**)malloc( nargin*sizeof(PyArrayIterObject*) ); 
  thisone->gsl_argin = (gsl_vector**)malloc( nargin*sizeof(gsl_vector*) ); 
  for( i=0; i<nargin; i++ ) {
    thisone->PyIter_argin[i] = (PyArrayIterObject*)PyArray_IterAllButAxis( (PyObject*)argin[i], &axis );
    thisone->gsl_argin[i] = _gsl_vector_fromPyArrayIter( thisone->PyIter_argin[i], axis );
  }
  
  return thisone;

#endif
}
