#include <Python.h>
#include <numpy/arrayobject.h>

#include <stdlib.h>

#ifndef FFF_ROUTINES_H
#define FFF_ROUTINES_H

#include "fff_types.h"

#ifndef NPY_VERSION
#define npy_intp intp
#define NPY_OWNDATA OWNDATA
#define NPY_CONTIGUOUS CONTIGUOUS
#define NPY_BEHAVED BEHAVED_FLAGS
#endif

/* This is a dummy constructor of an 0-sized vector */
#define fffPyZeroLONG()	(PyArrayObject*)PyArray_SimpleNew(1,(npy_intp*)"\0\0\0\0", PyArray_LONG);
#define fffPyZeroDOUBLE() (PyArrayObject*)PyArray_SimpleNew(1,(npy_intp*)"\0\0\0\0",PyArray_DOUBLE);

/*!  
  \struct fff_func_iterator 
  \brief Structure to iterate a vector-based function over arbitrary
  N-dimensional arrays

  This structure, which is close in spirit to numpy's universal
  functions, makes it easy to implement a C function that iterates
  over several numpy arrays simultaneously and calls a lower-level
  function whose arguments (in and out) are lists of gsl_vectors. Both
  input and output numpy arrays must have the same dimension, except
  in the direction defining the orientation of the gsl_vectors.
*/ 
typedef struct {
  npy_intp nargout; /*!< number of output numpy arrays */
  npy_intp nargin; /*!< number of input numpy arrays */
  npy_intp axis; /*!< gsl_vector direction */
  npy_intp nitems; /* number of iterations required */ 
  PyArrayIterObject** PyIter_argout; /* list of output PyArrays */ 
  PyArrayIterObject** PyIter_argin; /* list of input PyArrays */ 
  gsl_vector** gsl_argout; /* list of output gsl_vectors */ 
  gsl_vector** gsl_argin; /* list of input gsl_vectors */ 
} fff_func_iterator; 

/*!
   \brief Constructor for the \c fff_func_iterator structure
   \param nargout number of output numpy arrays
   \param argout list of output numpy arrays 
   \param nargin number of input numpy arrays
   \param argin list of input numpy arrays 
   \param axis vector orientation
   
   This constructor assumes that all PyArrays involved are already
   allocated, including output arrays. Moreover, they must have the
   same dimensions except in the direction given by \c axis.
*/ 
fff_func_iterator* fff_func_iterator_new( npy_intp nargout, PyArrayObject** argout, 
					  npy_intp nargin, PyArrayObject** argin,
					  npy_intp axis ); 

/*! 
   \brief Iterate a C function over several PyArrays
   \param thisone existing iterator structure 
   \param func pointer to the vector-based function to iterate
   \param par auxiliary function parameters
   
   \c func is intended to be a simple gateway function to an existing
   \a fff function. It should conform with the syntax:

   \code
   void func( gsl_vector** argout, gsl_vector** argin, void* par )
   \endcode 

   where \c argout is the list of output gsl_vectors, \c argin is the
   list of input gsl_vectors, and \c par is typically a reference to a
   pre-allocated structure in order to avoid memory allocations on
   each iteration. It is left to the developer to determine whether \c
   func assumes the correct number of input and output
   arguments. Hence, it is strongly recommended that such a gateway
   function be defined static in a Python module C file.
*/ 
void fff_func_iterator_eval( fff_func_iterator* thisone, 
			     void (*func)( gsl_vector**, gsl_vector**, void* ), 
			     void* par ); 

/*! 
   \brief Reset a \c fff_func_iterator before new loop 
   \param thisone existing iterator structure 

   This is useful when several loops over arrays are to be performed
   in the same C function.
*/ 
void fff_func_iterator_reset( fff_func_iterator* thisone ); 

/*! 
   \brief Destructor for the \c fff_func_iterator structure
   \param thisone iterator to delete

   Use after \c fff_func_iterator_new. 
*/ 
void fff_func_iterator_delete( fff_func_iterator* thisone ); 


double fff_quantile_from_temp_data( gsl_vector* x, double r, int interp );
static void _quantile( gsl_vector** argout, gsl_vector** argin, void* par );



static void _gsl_vector_toPyArrayIter_update( const gsl_vector* y, PyArrayIterObject* it, npy_intp axis );
static gsl_vector* _gsl_vector_fromPyArrayIter( const PyArrayIterObject* it, npy_intp axis );
static void _gsl_vector_fromPyArrayIter_update( gsl_vector* y, const PyArrayIterObject* it, npy_intp axis );
static void _gsl_vector_copy_fromPtr( gsl_vector* y, const char* data, npy_intp stride, int type ) ;
gsl_vector* _gsl_vector_fromPtr( const char* data, npy_intp dim, npy_intp stride, int type, int isaligned );
#endif
//----------------------- END OF FILE fff_routines_h ------------------------
