#include <fffPy.h>

#include <fff_macro.h>
#include <fff_routines.h>


#define PY_ARRAY_UNIQUE_SYMBOL PyArray_API

#define DEFAULT_AXIS 0

/* Code pour la creation du module */ 
static char quantile_doc[] = 
" T = quantile(data, ratio=.5, interp=False, axis=0)\n\
Comments to follow";
static char mahalanobis_doc[] = 
" d2 = mahalanobis(data, vardata)\n\
data and vardata assumed C contiguous. More comments to follow";
static char module_doc[] = 
" Basic statistical routines.\n\
Author: Alexis Roche (CEA, Orsay, France), 2004-2007.";


/* 
   Median
*/


/* Auxiliary struct for permutations */ 
typedef struct{
  gsl_vector* tmp; 
  double ratio;
  int interp; 
} quantile_struct;

static void _quantile( gsl_vector** argout, gsl_vector** argin, void* par )
{
  quantile_struct* Params = (quantile_struct*) par;
  gsl_vector_memcpy( Params->tmp, argin[0] ); 
  argout[0]->data[0] =  fff_quantile_from_temp_data( Params->tmp, 
						     Params->ratio, 
						     Params->interp ); 
  return; 
} 


#define DEFAULT_RATIO 0.5
#define DEFAULT_INTERP 0 

static PyObject* quantile(PyObject* self, PyObject* args, PyObject* kw) 
{
  double ratio = DEFAULT_RATIO;
  int interp = DEFAULT_INTERP;
  int axis = DEFAULT_AXIS;
  PyArrayObject *argin[1], *argout[1]; 
  npy_intp ndim, dim_axis;
  npy_intp* dimAux; 
  fff_func_iterator* FuncIter; 
  quantile_struct* par = NULL; 
  static char *kwlist[] = {"data", "ratio", "interp", "axis", NULL}; 
  
  /* Parse input */ 
  if ( !PyArg_ParseTupleAndKeywords( args, kw, "O!|dii:quantile", kwlist, 
				     &PyArray_Type, argin, &ratio, &interp, &axis ) )
    Py_RETURN_NONE; 
  
  /* Get input dimensions */
  ndim = PyArray_NDIM( argin[0] ); 
  dim_axis = PyArray_DIM( argin[0], axis );
  dimAux = (npy_intp*) malloc( ndim*sizeof(npy_intp) );
  memcpy( (void*)dimAux, (void*)PyArray_DIMS( argin[0] ), ndim*sizeof(npy_intp) ); 
 
  /* Allocate output array and local structs */
  dimAux[axis] = 1;
  argout[0] = (PyArrayObject*) PyArray_SimpleNew( ndim, dimAux, NPY_DOUBLE ); 
  par = (quantile_struct*) malloc( sizeof(quantile_struct) ); 
  par->ratio = ratio; 
  par->interp = interp; 
  par->tmp = gsl_vector_alloc( dim_axis ); 
  FuncIter = fff_func_iterator_new( 1, argout, 1, argin, axis );

  /* Compute statistics from non-randomized data */
  fff_func_iterator_eval( FuncIter, &_quantile, (void*)par ); 
  
  /* Free memory */ 
  free( dimAux ); 
  gsl_vector_free( par->tmp ); 
  free( par );
  fff_func_iterator_delete( FuncIter ); 
  
  /* Output */
  return (PyObject*)argout[0]; 
}



typedef struct{
  gsl_vector* work; 
  gsl_vector* S;
} mahalanobis_struct;


static void _mahalanobis( gsl_vector** argout, gsl_vector** argin, void* par )
{
  mahalanobis_struct* Params = (mahalanobis_struct*) par; 
  size_t n = argin[0]->size; 
  gsl_matrix_view Sview; 

  gsl_vector_memcpy(Params->S, argin[1]); /* par->S contiguous */  
  Sview = gsl_matrix_view_array(Params->S->data, n, n); 
  argout[0]->data[0] = fff_mahalanobis(argin[0], &Sview.matrix, Params->work);

  return; 
} 


static PyObject* mahalanobis(PyObject* self, PyObject* args, PyObject* kw) 
{
  PyArrayObject *argin[2], *argout[1];
  fff_func_iterator* FuncIter; 
  static char *kwlist[] = {"data", "vardata", NULL}; 
  npy_intp ndim, n;
  npy_intp* dimAux;  
  mahalanobis_struct par; 

  /* Parse input */ 
  if ( !PyArg_ParseTupleAndKeywords( args, kw, "O!O!:mahalanobis", kwlist, 
				     &PyArray_Type, &(argin[0]),
				     &PyArray_Type, &(argin[1]) ) )
    Py_RETURN_NONE; 
  
  
  /* Dimensions */
  n = PyArray_DIM( argin[0], 0 ); 
  ndim = PyArray_NDIM( argin[0] ); 
  
  /* Allocate output array and local structs */
  dimAux = (npy_intp*) malloc( ndim*sizeof(npy_intp) );
  memcpy( (void*)dimAux, (void*)PyArray_DIMS( argin[0] ), ndim*sizeof(npy_intp) ); 
  dimAux[0] = 1;
  par.work = gsl_vector_alloc( n ); 
  par.S = gsl_vector_alloc( n*n ); 
  argout[0] = (PyArrayObject*) PyArray_SimpleNew( ndim, dimAux, NPY_DOUBLE ); 
  FuncIter = fff_func_iterator_new( 1, argout, 2, argin, 0 );

  /* Do ze job */ 
  fff_func_iterator_eval( FuncIter, &_mahalanobis, (void*)&par ); 

  /* Free memory */ 
  free( dimAux ); 
  gsl_vector_free( par.work ); 
  gsl_vector_free( par.S ); 
  fff_func_iterator_delete( FuncIter ); 
  
  /* Output */
  return (PyObject*)argout[0]; 
}

 



static PyMethodDef module_methods[] = {
  {"quantile", (PyCFunction)quantile, METH_KEYWORDS, quantile_doc},
  {"mahalanobis", (PyCFunction)mahalanobis, METH_KEYWORDS, mahalanobis_doc},
  {NULL, NULL, 0, NULL}
};


void initroutines(void)
{
  Py_InitModule3("routines", module_methods, module_doc);
  fffPy_import_array();
  import_array();   /* required NumPy initialization */

  return; 
}
